package edu.toronto.cs.ctl;

import java.util.*;

/**
 ** This is an abstract CTL node.
 **/
public abstract class CTLAbstractNode implements CTLNode
{

  protected CTLNode right, left;
  protected List fairness;

  /**
   ** Create an CTLAbstractNode with no children.
   **/
  protected CTLAbstractNode ()
  {
    right = null;
    left = null;
  }

  /**
   ** Create an CTLAbstractNode with one child.
   **/
  protected CTLAbstractNode (CTLNode onlychild)
  {
    left = null;
    right = onlychild;
  }

  /**
   ** Create an CTLAbstractNode with two children.
   **/
  protected CTLAbstractNode (CTLNode l, CTLNode r)
  {
    left = l;
    right = r;
  }

  /**
   ** Retrieves the fairness of the CTLNode.
   **
   ** @return fairness for this CTL.
   **/
  public List getFairness ()
  {
    List result = new LinkedList ();
    result.addAll (fairness);
    return result;
  }

  /**
   ** Retrieves the left child of the CTLNode.
   **
   ** @return left child.
   **/
  public CTLNode getLeft ()
  {
    return left;
  }

  /**
   ** Retrieves the right child of the CTLNode.
   **
   ** @return right child.
   **/
  public CTLNode getRight ()
  {
    return right;
  }

  /**
   ** Changes the fairness of the CTLNode.
   **
   ** @return fairness for this CTL.
   **/
  protected void setFairness (List c)
  {
    fairness = new LinkedList ();
    fairness.addAll (c);
  }
    
  /**
   ** Changes the left child of the CTLNode.
   **/
  protected void setLeft (CTLNode l)
  {
    left = l;
  }

  /**
   ** Changes the right child of the CTLNode.
   **/
  protected void setRight (CTLNode r)
  {
    right = r;
  }

  /**
   ** CTL operation - conjunction.
   **/
  public CTLAndNode and (CTLNode psi)
  {
    return CTLFactory.createCTLAndNode (this, psi);
  }

  /**
   ** CTL operation - disjunction.
   **/
  public CTLOrNode or (CTLNode psi)
  {
    return CTLFactory.createCTLOrNode (this, psi);
  }

  /**
   ** CTL operation - negation.
   **/
  public CTLNegNode neg ()
  {
    return CTLFactory.createCTLNegNode (this);
  }

  /**
   ** CTL operation - implication.
   **/
  public CTLImplNode implies (CTLNode psi)
  {
    return CTLFactory.createCTLImplNode (this, psi);
  }

  /**
   ** CTL operation - backward implication.
   **/
  public CTLImplNode impliedBy (CTLNode psi)
  {
    return CTLFactory.createCTLImplNode (psi, this);
  }

  /**
   ** CTL operation - equivalence.
   **/
  public CTLIffNode iff (CTLNode psi)
  {
    return CTLFactory.createCTLIffNode (this, psi);
  }

  /**
   ** CTL operation - EX.
   **/
  public CTLEXNode ex ()
  {
    return CTLFactory.createCTLEXNode (this);
  }

  /**
   ** CTL operation - AX.
   **/
  public CTLAXNode ax ()
  {
    return CTLFactory.createCTLAXNode (this);
  }

  /**
   ** CTL operation - EF.
   **/
  public CTLEFNode ef ()
  {
    return CTLFactory.createCTLEFNode (this);
  }

  /**
   ** CTL operation - AF.
   **/
  public CTLAFNode af ()
  {
    return CTLFactory.createCTLAFNode (this);
  }

  /**
   ** CTL operation - EU ( E[this U psi] ).
   **/
  public CTLEUNode eu (CTLNode psi)
  {
    return CTLFactory.createCTLEUNode (this, psi);
  }

  /**
   ** CTL operation - bounded EU ( E[this Ui psi] ).
   **/
  public CTLEUiNode eu (int i, CTLNode psi)
  {
    return CTLFactory.createCTLEUiNode (this, i, psi);
  }

  /**
   ** CTL operation - AU ( A[this U psi] ).
   **/
  public CTLAUNode au (CTLNode psi)
  {
    return CTLFactory.createCTLAUNode (this, psi);
  }

  /**
   ** CTL operation - bounded AU ( A[this Ui psi] ).
   **/
  public CTLAUiNode au (int i, CTLNode psi)
  {
    return CTLFactory.createCTLAUiNode (this, i, psi);
  }

  /**
   ** CTL operation - EG.
   **/
  public CTLEGNode eg ()
  {
    return CTLFactory.createCTLEGNode (this);
  }

  /**
   ** CTL operation - AG.
   **/
  public CTLAGNode ag ()
  {
    return CTLFactory.createCTLAGNode (this);
  }

  /**
   ** CTL operation - equals
   **/
  public CTLEqualsNode eq (CTLNode psi)
  {
    return CTLFactory.createCTLEqualsNode (this, psi);
  }

  /**
   ** CTL operation - under
   **/
  public CTLUnderNode under (CTLNode psi)
  {
    return CTLFactory.createCTLUnderNode (this, psi);
  }

  /**
   ** CTL operation - over
   **/
  public CTLOverNode over (CTLNode psi)
  {
    return CTLFactory.createCTLOverNode (this, psi);
  }

  /**
   ** Overrides the Object's equals method.
   **/
  public boolean equals (Object o)
  {
    return this == o;
  }

  /**
   ** This is used to determine uniqueness.
   **/
  public String toFairString ()
  {
    String result;
    
    result = this.toString() + fairness;
    if (getLeft () != null) result = result +
			      ((CTLAbstractNode) getLeft()).fairness;
    if (getRight () != null) result = result +
			       ((CTLAbstractNode) getRight()).fairness;
    return result;
  }

  public void accept (CTLVisitor v, Object stateinfo)
  {
    v.visitAbstractNode (this, stateinfo);
  }

}

