package edu.toronto.cs.cudd;

import edu.toronto.cs.util.Logger;
/***
 *** Wrapper for CUDD
 ***/
public class Cudd
{
  static Logger log = new Logger (Cudd.class.getName ());
  
  // -- pointer to the Cudd manager
  int mgrPtr;

  BDD bddOne = null;
  BDD bddZero = null;

  static 
  {
    System.loadLibrary ("jcudd");
  }
  
  public Cudd (int initVars)
  {
    log.debug ("Intializing");
    init (initVars);
    log.debug ("Done initialization");
  }
  
  public Cudd ()
  {
    this (0);
  }

  // -- initialize the mgrPtr
  native void init (int initVars);

  private native synchronized int cuddBddOnePtr ();
  private native synchronized int cuddBddZeroPtr ();

  public int bddOnePtr ()
  {
    //return bddOne ().getNodePtr ();
    return cuddBddOnePtr ();
  }
  public int bddZeroPtr ()
  {
    //return bddZero ().getNodePtr ();
    return cuddBddZeroPtr ();
  }
  
  

  public native synchronized int bddVarPtr (int index);  

  public native synchronized int bddAndPtr (int a, int b);
  public native synchronized int bddOrPtr (int a, int b);
  public native synchronized int bddNotPtr (int a);
  public native synchronized int bddConstrainPtr (int f, int c);
  public native synchronized int bddPermutePtr (int a, int[] map);

  public native synchronized int existAbstract (int ptr, int cube);
  public native synchronized int forallAbstract (int ptr, int cube);  
  public native synchronized void ref (int ptr);
  public native synchronized void deref (int ptr);
  public native synchronized void recursiveDeref (int ptr);
  
  public native int refCount (int ptr);

  public native int checkZeroRef ();
  public native synchronized void quit ();

  // -- outputs info on stdout
  public native void info ();

  public native synchronized void reorder (int minsize);
  public native synchronized int dagSize (int ptr);
  public native synchronized int sharingSize (int[] ptrs);


  // -- transfers the ptr from current DdManager to the dst one 
  // -- scaling the variables by a factor
  public native int transferPtr (Cudd dstCudd, int ptr, int factor);

  /***
   *** Computes a function h(x) = f(x) >= g(x)
   *** using h(x) = f(x) \/ !g(x)
   ***/
  public int geqPtr (int f, int g)
  {
    int notG = bddNotPtr (g);
    int h = bddOrPtr (f, notG);
    // -- free !g
    recursiveDeref (notG);
    return h;
  }
  
  public int leqPtr (int f, int g)
  {
    // -- f <= g is just g >= f
    return geqPtr (g, f);
  }
  
  public int eqPtr (int f, int g)
  {
    int leq = leqPtr (f, g);
    int geq = geqPtr (f, g);
    
    int h = bddAndPtr (leq, geq);
    recursiveDeref (leq);
    recursiveDeref (geq);
    return h;
  }
  

  protected void finalize () throws Throwable
  {
    super.finalize ();
    if (mgrPtr == 0) return;


    dead = true;

    int retval = checkZeroRef ();
    // -- XXX Something is very strange but this seems to fix it
    // -- XXX Read about finalize and see how it is called, what 
    // -- XXX threads call it, etc. It seems that finalization happens
    // -- XXX in random order but multiple threads, and thus 
    // -- XXX DD nodes are finalized after or during the time the manager
    // -- XXX is finalized and this causes problems
    System.out.println ("\t entering wait for glory");
    for (int i = 0; checkZeroRef () != 0 && i < 10; i++)
      //while (checkZeroRef () != 0)
      Thread.currentThread ().sleep (100);
    System.out.println ("\t done waiting for glory");
    //System.exit (0);
    
    if (retval != 0)
      log.warn (retval + " unexpected non-zero reference count");
    else
      log.debug ("All went well");
    quit ();
    mgrPtr = 0;
    System.out.println ("CUDD: done finalizing");
  }
  
  boolean dead = false;
  public boolean isDead ()
  {
    return dead;
  }
  
  
  public int restrictVar (int ptr, int var, boolean val)
  {
    int varBdd = bddVarPtr (var);
    
    if (!val)
      {
	int temp = bddNotPtr (varBdd);
	recursiveDeref (varBdd);
	varBdd = temp;
      }

    int result = bddConstrainPtr (ptr, varBdd);
    recursiveDeref (varBdd);
    return result;
  }
  
  
  // -- returns a bdd corresponding to 'index'
  public BDD bddVar (int index)
  {
    return new BDD (this, bddVarPtr (index));
  }
  
  public BDD bddOne ()
  {
    if (bddOne == null)
      bddOne = new BDD (this, cuddBddOnePtr ());
    return bddOne;
  }
  
  public BDD bddZero ()
  {
    if (bddZero == null)
      bddZero = new BDD (this, cuddBddZeroPtr ());
    return bddZero;
  }
  

  public String toString ()
  {
    return String.valueOf (mgrPtr);
  }
  
  public static void main (String[] args)
  {
    Cudd cudd = new Cudd ();
    System.out.println ("Got: " + cudd);
    System.out.println ("1: " + cudd.bddOnePtr ());
    System.out.println ("0: " + cudd.bddZeroPtr ());
    System.out.println ("1 * 0: " + 
			cudd.bddAndPtr 
			(cudd.bddOnePtr (), 
			 cudd.bddZeroPtr ()));
    System.out.println ("1 + 0: " + 
			cudd.bddOrPtr 
			(cudd.bddOnePtr (), 
			 cudd.bddZeroPtr ()));
    
  }

  public boolean equals (Object o)
  {
    return o != null && o instanceof Cudd && equals ((Cudd)o);    
  }

  public boolean equals (Cudd o)
  {
    return getMgrPtr () == o.getMgrPtr ();
  }
  
  int getMgrPtr ()
  {
    return mgrPtr;
  }
  
  

  public static class BDD
  {
    Cudd manager;
    int nodePtr;
    
    protected BDD (Cudd _manager, int _nodePtr)
    {
      manager = _manager;
      nodePtr = _nodePtr;
    }

    public BDD bddAnd (BDD bdd)
    {
      return 
	new BDD (manager, manager.bddAndPtr (nodePtr, bdd.getNodePtr ()));
    }
    public BDD bddOr (BDD bdd)
    {
      return new BDD (manager, manager.bddOrPtr (nodePtr, bdd.getNodePtr ()));
    }
    public BDD bddNot ()
    {
      return new BDD (manager, manager.bddNotPtr (nodePtr));
    }
    public BDD constrain (BDD c)
    {
      return 
	new BDD (manager, manager.bddConstrainPtr (nodePtr, c.getNodePtr ()));
    }
    
    public String toString ()
    {
      return manager + "@" + nodePtr;
    }
    
    
    Cudd getManager ()
    {
      return manager;
    }
    
    int getNodePtr ()
    {
      return nodePtr;
    }
    
    protected void finalize () throws Throwable
    {
      super.finalize ();
      if (nodePtr != 0)
	manager.recursiveDeref (nodePtr);
    }
  

    public boolean equals (Object o)
    {
      return o != null && o instanceof BDD && equals ((BDD)o);
    }
    
    public boolean equals (BDD o)
    {
      return getManager ().equals (o.getManager ()) &&
	getNodePtr () == o.getNodePtr ();
    }
    
    
  }
  
	
  
}
