package edu.toronto.cs.experiments;

import java.util.*;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.util.*;


public class MvNoiseExperiment2
{

  // -- number of cubes to alter
  int noiseLevel;
  FiniteLattice lattice;

  int top;
  int bot;

  ChaosIterator varDomain;

  ExperimentOutput output;
  

  String ddFileName;


  ChaosIterator valDomain;
  
  public MvNoiseExperiment2 (FiniteLattice _lattice, 
			      ExperimentOutput _output, 
			      String _ddFileName,
			      int _noiseLevel)
  {
    
    lattice = _lattice;
    output = _output;
    ddFileName = _ddFileName;
    noiseLevel = _noiseLevel;

    top = lattice.idtimesIdx ();
    bot = lattice.idplusIdx ();
    
    varDomain = new UniformChaosIterator (new int[] {top, bot});
    
    //valDomain = buildValDomain ();
    
  }
  
  // -- a pool of mv-values for transitions
  ChaosIterator buildValDomain ()
  {
    // -- assign only non join-irreducible elements
    int[] joinIrr = lattice.getJoinIrreducibles ();
    Set joinIrrSet = new HashSet ();
    for (int i = 0; i < joinIrr.length; i++)
      joinIrrSet.add (new Integer (joinIrr [i]));

    // -- add \top and \bot since we want to exclude them as well
    joinIrrSet.add (new Integer (lattice.idtimesIdx ()));
    joinIrrSet.add (new Integer (lattice.idplusIdx ()));

    Set valDomainSet = new HashSet ();
    
    for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
      {
	Integer val = new Integer (it.nextInt ());
	if (joinIrrSet.contains (val)) continue;
	valDomainSet.add (val);
      }
    

    int[] valDomainArray = new int [valDomainSet.size ()];
    int count = 0;
    
    for (Iterator it = valDomainSet.iterator (); it.hasNext ();)
      valDomainArray [count++] = ((Integer)it.next ()).intValue ();
    
    return new UniformChaosIterator (valDomainArray);
  }
  

  public void runExperiment ()
  {

    // -- create our manager XXX ignore arrity
    MDDManager manager = new MDDMgr (0, lattice);
    

    // -- get our join irreducibles
    int[] joinIrr = lattice.getJoinIrreducibles ();
    
    MDDNode[] mbtddNodes = new MDDNode [joinIrr.length];
    
    // -- first view point comes from a file
    MDDNode bddNode = DDUnDumper.readBDDSafe (manager, ddFileName);

    // -- priorities the original diagram
    MDDNode mddNode = introduceMvNoise (manager, bddNode);

//     Assert.assert (mddNode != bddNode);

    // -- mdd sample is ready
    output.addSample ("mdd", manager.dagSize (mddNode));
    

    // -- create mbtddNodes now
    for (int i = 0; i < mbtddNodes.length; i++)
      {
	//System.out.println ("Perfoming a cut for: " + joinIrr [i]);
	// -- Let f be an mdd, and let h be its mbtdd encoding, then
	// -- h_i = f \above i
	mbtddNodes [i] = manager.Apply (mddNode, 
					manager.getConstant (joinIrr [i]),
					FiniteLattice.ABOVE);
      }
    

//     Assert.assert (mbtddNodes [0] != mddNode);
    
//     Assert.assert (mbtddNodes [0] != mbtddNodes [1]);
      

    // -- we are ready for the 'mbtdd' case
    output.addSample ("mbtdd", manager.sharedSize (mbtddNodes));
    output.flush ();
  }
  

			   
  /*** here we get a random value that is not \top or \bot ***/
  public int getRandomNegativeValue ()
  {
    //return valDomain.nextInt ();
    int top = lattice.idtimesIdx ();
    int bot = lattice.idplusIdx ();


    // -- find the minimum
    int lo = top < bot ? top : bot;
    
    // -- always return the same value
    return lo + 1;
  }
  
  // -- introduce mv-noise into a decision diagram
  public  MDDNode introduceMvNoise (MDDManager manager, MDDNode node)
  {
    // -- given a decision diagram, and an FTO with the middle value M
    // -- we want to map some \top cubes to the interval (\top, M]
    // -- and some \bot cubes to the interval [M, \top)
    MDDNode positiveCubes = manager.getZero ();
    MDDNode negativeCubes = manager.getZero ();
    
    // -- find cubes we will change
    for (int i = 0; i < noiseLevel; i++)
      {
	// -- get a random point
	MDDPoint point = manager.getRandomPoint (node, varDomain);
	if (point.getValue () == top)
	  {

	    // -- got a positive cube that we want to augment
	    
	    // -- get a random negative value i.e. in interval [M, \bot)
	    // -- since we will negate everything later on anyways
	    int val = getRandomNegativeValue ();

	    MDDNode pointNode = point.getNode ();
	    
	    // -- set the value
	    pointNode = manager.setEqualTo (pointNode, val);
	    
	    
	    // -- add to the rest of our cubes
	    positiveCubes = manager.Apply (positiveCubes, pointNode, 
					   FiniteLattice.JOIN);
	  }
	else
	  {
	    int val = getRandomNegativeValue ();
	    MDDNode pointNode = point.getNode ();
	    pointNode = manager.setEqualTo (pointNode, val);
	    
	    negativeCubes = manager.Apply (negativeCubes, pointNode,
					   FiniteLattice.JOIN);
	  }
	
      }
    

    MDDNode result = node;

    System.out.println ("Merging all negative cubes...");
    // -- change all of the negative cubes    
    result = manager.Apply (result, negativeCubes, 
			    FiniteLattice.JOIN);
    
    System.out.println ("Merging all positive cubes....");
    // -- change all positive cubes
    result = manager.Apply 
      (result, manager.Apply (positiveCubes, FiniteLattice.NEG),
       FiniteLattice.MEET);
    return result;
  }
  

  public static void main (String[] args) throws Exception
  {
    // -- intialize seed so that we could re-run this
    Chaos.seed (0);
    
    String ddFileName = args [0];
    String outputDir = args [1];
    String experimentName = args [2];


    String latticeName = args [3];
    int noiseLevel = Integer.parseInt (args [4]);
    
    
    ExperimentOutput output = 
      new ExperimentOutput (outputDir, experimentName);
    FiniteLattice lattice;

    
    
    lattice = Lattices.getLatticeByName (latticeName);
    System.out.println ("Getting a lattice: " + latticeName);

    for (int i = 0; i < 100; i++)
      {
	System.out.println ("Getting sample: " + i);
	MvNoiseExperiment2 experiment = 
	  new MvNoiseExperiment2 (lattice, 
				  output, 
				  ddFileName,
				  noiseLevel);
	System.out.println ("Running experiment....");
	experiment.runExperiment ();
	System.out.println ("Done");
      }
    
  }
  
}
