package edu.toronto.cs.gui;

import edu.toronto.cs.modelchecker.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.tlq.*;
import edu.toronto.cs.smv.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.keg.*;
import edu.toronto.cs.util.dgraph.*;

import javax.swing.*;
import javax.swing.event.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;

/**
 ** The GUI for our Model-Checker.
 **/
public class XCGui extends JFrame
{
  // name of the application
  final static String GUINAME = "XChek";

  final static String LOADERFILE = ".xbel_modelloaders";

  InputPanel ctlin;
  OutputPanel output;
  OptionsPanel optionspanel;
  XCMenu menu;
  XCToolbar toolbar;
  ModelInfoFrame modelInfo = null;
  LatticeDisplay lattice = null;
  
  // maps the name of the action to the appropriate Action object
  ActionMap actionMap;
  
  // options for the XC
  Map options = new HashMap ();

  MCManager mcman = null;

  // the macro expander for handling state names
  //  StringMacroManager macro;
  CTLMacroManager macroManager;

  //  PlaceholderManager2 placeHolderManager = null;

  // -- enumerated types that we know about
  EnumeratedTypes enumTypes = EnumeratedTypes.EMPTY_TYPES;

  File currentDir = new File (System.getProperty ("user.dir"));
  JFileChooser filechooser = new JFileChooser (currentDir);

  Properties loaderProperties = new Properties ();
  Properties loaderClassNames = new Properties ();
  java.util.List loaderClasses = new ArrayList ();

  final private String version = "0.7";

  public XCGui ()
  {
    super (GUINAME);
    init ();
  }


  private void initApp ()
  {
    loadLoaders ();
  }
  

  private void init ()
  {
    
    initApp ();

    initActionMap();
    
    menu = new XCMenu (actionMap);
    toolbar = new XCToolbar (actionMap);

    setJMenuBar (menu);

    // make the layout a bit nicer (more space between everything)
    JPanel all = new JPanel ();
    all.setLayout (new BorderLayout ());
    ((BorderLayout) all.getLayout ()).setHgap (StandardFiller.getSize ());
    ((BorderLayout) all.getLayout ()).setVgap (StandardFiller.getSize ());    
    all.setBorder (StandardFiller.makeWideEmptyBorder ());

    // now add all of the internals
    all.add (createCTLInteraction (),
	     BorderLayout.SOUTH);
    all.add (createOutputArea (),
	     BorderLayout.CENTER);
    all.add (createOptions (),
	     BorderLayout.EAST);
    all.add (toolbar,
	    BorderLayout.NORTH);

    optionspanel.setEnableTabs (false);

    getContentPane ().add (all);
    setDefaultCloseOperation (JFrame.EXIT_ON_CLOSE);

    // set up some model-checking options
    setBasicOptions ();
    if (options.get ("classprefix") == null) options.put ("classprefix", "");
  }  


  /**
   ** Initialize the set of actions and add them to actionMap
   **/
  private void initActionMap ()
  {

    actionMap = new ActionMap();
    
    Action action;

    // Open XML model action
    action = new AbstractAction ("Open xKripke (XML) model...", 
				 new ImageIcon (openXMLIcon))
                                 {
                                   public void actionPerformed (ActionEvent e)
	    			    {
				      loadXMLmodel();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Open XML model");
    actionMap.put ("Open XML model...", action);

    // Open BDD model actiong
    action = new AbstractAction ("Open BDD model...",
				 new ImageIcon (openBDDIcon))
                                 { 
                                   public void actionPerformed (ActionEvent e)
				    {
				      loadBDDmodel ();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Open BDD model");
    actionMap.put ("Open BDD model...", action);
  
    // Open a generic model action
    action = new AbstractAction ("Open from custom source...",
				 new ImageIcon (openBDDIcon))
                                 { 
                                   public void actionPerformed (ActionEvent e)
				    {
				      loadFromGeneric ();
				      //openModel ();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Open a generic model");
    actionMap.put ("Open a generic model...", action);
  
    action = new AbstractAction ("Open a model...",
				 new ImageIcon (openBDDIcon))
                                 { 
                                   public void actionPerformed (ActionEvent e)
				    {
				      openModel ();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Open a model...");
    actionMap.put ("Open a model...", action);

    // Load fairness conditions action
    action = new AbstractAction ("Load fairness conditions",
				 new ImageIcon (loadFairnessIcon))
                                 { 
                                   public void actionPerformed (ActionEvent e)
				    {
				      loadFairness ();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Load fairness conditions");
    action.setEnabled(false);
    actionMap.put ("Load fairness conditions", action);
  
    // Save fairness conditions action
    action = new AbstractAction ("Save fairness conditions",
				 new ImageIcon (saveFairnessIcon))
                                 { 
                                   public void actionPerformed (ActionEvent e)
				    {
				      saveFairness ();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Save fairness conditions");
    action.setEnabled(false);
    actionMap.put ("Save fairness conditions", action);

    // Load CTL history action
    action = new AbstractAction ("Load CTL history",
				 new ImageIcon (loadCTLIcon))
                                 { 
                                   public void actionPerformed (ActionEvent e)
				    {
				      loadCTLHistory ();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Load CTL history");
    action.setEnabled(false);
    actionMap.put ("Load CTL history", action);
  
    // Save fairness conditions action
    action = new AbstractAction ("Save CTL history",
				 new ImageIcon (saveCTLIcon))
                                 { 
                                   public void actionPerformed (ActionEvent e)
				    {
				      saveCTLHistory ();
				    }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Save CTL history");
    action.setEnabled(false);
    actionMap.put ("Save CTL history", action);

    // Quit action
    action = new AbstractAction ("Quit", new ImageIcon (quitIcon))
                                 {
				   public void actionPerformed (ActionEvent e)
				   {
				     quit();
				   }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Quit");
    actionMap.put ("Quit", action);
    
    // Show model info action
    action = new AbstractAction ("Model info", new ImageIcon (modelInfoIcon))
                                 {
				   public void actionPerformed (ActionEvent e)
				   {
				     showModelInfo ();
				   }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Display model info");
    actionMap.put ("Model info", action);

    // Show lattice info action
    action = new AbstractAction ("Lattice info", new ImageIcon (latticeIcon))
                                 {
				   public void actionPerformed (ActionEvent e)
				   {
				     showLatticeDisplay ();
				   }
                                 };
    action.putValue(Action.SHORT_DESCRIPTION, "Display lattice info");
    action.setEnabled(false);
    actionMap.put ("Lattice info", action);


    // Show variables action
    action = new AbstractAction ("Show variables", new ImageIcon (varIcon))
                                 {
				   public void actionPerformed (ActionEvent e)
				   {
				     showVariables ();
				   }
                                 };
    
    actionMap.put ("Show variables", action);

    // Show general help
    action = new AbstractAction ("Help", null)
                                 {
				   public void actionPerformed (ActionEvent e)
				   {
				     showHelp ();
				   }
                                 };
    
    actionMap.put ("Help", action);
    // About
    action = new AbstractAction ("About", null)
                                 {
				   public void actionPerformed (ActionEvent e)
				   {
				     showAbout ();
				   }
                                 };
    
    actionMap.put ("About", action);
  }

  /**
   ** Creates the part of the GUI that deals with inputting a CTL
   ** formula and running the model-checker on it ("run" button).
   **/
  private Component createCTLInteraction ()
  {
    ctlin = new InputPanel ();
    GUIutil.setEnabled (ctlin, false); // disable CTL input by default
    ctlin.addRunActionListener (new ActionListener ()
				{
				  public void actionPerformed (ActionEvent e)
				    {
				      runOnCTL ();
				    }
				});
    ctlin.addCTLKeyListener (new KeyAdapter ()
			     {
			       public void keyPressed (KeyEvent e)
				 {
				   if (e.getKeyCode () == KeyEvent.VK_ENTER)
				     runOnCTL ();
				 }
			     });
    return ctlin;
  }

  private Component createOutputArea ()
  {
    output = new OutputPanel ();
    return output;
  }

  /**
   ** Creates model-checking options - fairness, macroes.
   **/
  private Component createOptions ()
  {
    optionspanel = new OptionsPanel ();
    return optionspanel;
  }

  public static void main (String [] args)
  {
    try
      {
	UIManager.setLookAndFeel
	  //("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
	  //("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
	  (UIManager.getCrossPlatformLookAndFeelClassName ());
      }
    catch (Exception e) { }

    //Create the top-level container and add contents to it.
    XCGui app = new XCGui ();

    //Finish setting up the frame, and show it.
    app.pack();
    app.setVisible (true);
    app.options = GXClapi.parseOpts (args);
    if ( !(((java.util.List)app.options.get ("input")).get (0).
	   toString ().equals ("")) )
      app.finishLoadingModel ();
  }

  /**
   ** Reads the CTL formula to be used in the current model-checker run.
   **/
  public void runOnCTL ()
  {
    // read the ctl from the combo box
    CTLNode ctl = ctlin.getCTL (mcman.getMC ().getLattice (),
				optionspanel.getFairness ());

    if (ctl != null)
      checkCTL (ctl);
  }

  /**
   ** Sets the default core options for XCheck.
   **/
  private void setBasicOptions ()
  {
    options.put ("kegmode", new Boolean (false));
    options.put ("verbose", new Integer (0));
    options.put ("mvset", null);
    options.put ("initstate", null);
    options.put ("statenames", new Boolean (false));
    options.put ("smv", new Boolean (false));
    options.put ("bdd", new Boolean (false));
    options.put ("heuristic", null);
    options.put ("prefix", "KEG");
    options.put ("output", "out.daVinci");
    //    options.put ("input", Arrays.asList (inputValues));
    options.put ("expansion", "a");    
    options.put ("model name", "");
  }

  /**
   ** load XML model
   **/
  private void loadXMLmodel ()
  {
    // -- kill prev. model if it exists
    boolean openNewModel = true;
    if (mcman != null) openNewModel = closeModel ();

    if (openNewModel)
      {
	filechooser.setDialogTitle ("Open XML model");
	int result = filechooser.showOpenDialog (this);
	if (result == JFileChooser.APPROVE_OPTION)
	  {
	    // set the input option
	    setBasicOptions ();
	    java.util.List inputfile = Arrays.asList
	      (new String [] {filechooser.getSelectedFile ().
			      getAbsolutePath ()});
	    options.put ("input", inputfile);
	    // set-up the xDD type to be used
	    selectxDD ();
	    finishLoadingModel ();
	  }
      }
    else output.append ("Cancelled\n");
  }  

  /**
   ** load BDD model
   **/
  private void loadBDDmodel ()
  {
    // -- kill prev. model if it exists
    boolean openNewModel = true;
    if (mcman != null) openNewModel = closeModel ();

    if (openNewModel)
      {
	OpenBDDPanel fc = new OpenBDDPanel (filechooser);
	fc.pack ();
	fc.setLocation ((int)((this.getBounds ().getWidth ()-fc.getSize ().getWidth ())/2+this.getBounds ().getX ()), (int) ((this.getSize ().getHeight ()-fc.getBounds ().getHeight ())/2 + this.getBounds ().getY ()));
	fc.setVisible (true);
	String names [] = fc.getFileNames ();
	if (names != null)
	  {
	    // set up the input files option
	    setBasicOptions ();
	    String [] temp = new String [1 + names.length];
	    temp [0] = "edu.toronto.cs.smv.SMVBddModule";
	    for (int i = 0; i < names.length; i++) temp [i+1] = names [i];
	    java.util.List inputfiles = Arrays.asList (temp);
	
	    options.put ("smv", new Boolean (true));
	    options.put ("input", inputfiles);
	    System.out.println ("This is the inputfile list: "+inputfiles);
	    options.put ("mvset", "mdd");
	    finishLoadingModel ();
	  }
      }
    else output.append ("Cancelled\n");
  }

  private void loadFromGeneric ()
  {
    GenericSelectionList gsl = new GenericSelectionList 
      (loaderClasses, loaderClassNames, loaderProperties);
    gsl.setLocation ((int)((this.getBounds ().getWidth ()-gsl.getSize ().getWidth ())/2+this.getBounds ().getX ()), (int) ((this.getSize ().getHeight ()-gsl.getBounds ().getHeight ())/2 + this.getBounds ().getY ()));
    gsl.setVisible (true);
    String name = gsl.getSelection ();
    if (name != null)
      {
	setBasicOptions ();
	options.put ("smv", new Boolean (true));

	String allinput = loaderClassNames.get(name) + " " +
	  loaderProperties.get (name);
	
	System.out.println ("Your allinput is: "+allinput);
	
	// break it into tokens and dump it into an array
	StringTokenizer strtokens = new StringTokenizer (allinput);

	//	String [] input = new String [strtokens.countTokens ()]	
// 	for (int i = 0; i < input.length; i++)
// 	  input [i] = strtokens.nextToken ();
// 	java.util.List inputs = Arrays.asList (input);
	java.util.List input = new ArrayList ();
	int numTokens = strtokens.countTokens ();
	for (int i = 0; i < numTokens; i++)
	  {
	    input.add (strtokens.nextToken ());
	  }

	System.out.println ("This is the inputfile list: "+input);

	options.put ("input", input);
	options.put ("mvset", "mdd");

	selectxDD ();
	finishLoadingModel ();
      }    
  }


  class GenericSelectionList extends JDialog
  {
    boolean chose = false;
    java.util.List classes;
    Properties classNames;
    Properties classParams;
    JScrollPane listHolder;
    JList listOfOptions = null;
    ActionMap localActionMap = new ActionMap ();

    public GenericSelectionList 
      (java.util.List _classes, Properties _classNames, 
       Properties _classParams)
    {
      super ((Frame)null, "Choose custom type", true);
      //JDialog d = new JDialog ((Frame)null, "Choose Model Source", true);
      classes = _classes;
      classNames = _classNames;
      classParams = _classParams;
      JPanel allPanel = new JPanel ();
      allPanel.setLayout (new BoxLayout (allPanel, BoxLayout.Y_AXIS));
      
      JPanel listPanel = new JPanel ();
      listPanel.setLayout (new BoxLayout (listPanel, BoxLayout.Y_AXIS));

      JPanel buttonPanel = new JPanel ();
      buttonPanel.setLayout (new BoxLayout (buttonPanel, BoxLayout.X_AXIS));

      JPanel listControls = new JPanel ();
      listControls.setLayout (new BoxLayout 
			      (listControls, BoxLayout.X_AXIS));      

      JButton ok = new JButton (new okActionListener (this));
      localActionMap.put (ok.getAction ().getValue (Action.NAME), 
			  ok.getAction ());
      ok.setText ("Load");
      
      JButton edit = new JButton (new editAction (this));
      localActionMap.put (edit.getAction ().getValue (Action.NAME), 
			  edit.getAction ());
      edit.setText ("Edit");

      listOfOptions = new JList (new Vector (classes));
      listHolder = new JScrollPane (listOfOptions);
      listHolder.setVerticalScrollBarPolicy 
	(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

      JButton cancel = new JButton (new cancelActionListener (this));
      localActionMap.put (cancel.getAction ().getValue (Action.NAME), 
			  cancel.getAction ());
      cancel.setText ("Cancel");

      JButton delete = new JButton (new deleteAction (this));
      localActionMap.put (delete.getAction ().getValue (Action.NAME), 
			  delete.getAction ());

      JButton add = new JButton (new addAction (this));
      localActionMap.put (add.getAction ().getValue (Action.NAME),
			  add.getAction ());
      add.setText ("Add");

      ListSelectionListener l = new ListListener (this);
      listOfOptions.addListSelectionListener (l);

      listControls.add (edit);
      listControls.add (add);
      listControls.add (delete);      

      buttonPanel.add (cancel);
      buttonPanel.add (ok);

      listPanel.add (listHolder);
      listPanel.add (listControls);
      allPanel.add (listPanel);
      allPanel.add (buttonPanel);
      //allPanel.setLayout (new BoxLayout (allPanel, BoxLayout.Y_AXIS));
      allPanel.setBorder (StandardFiller.makeWideEmptyBorder ());
      this.getContentPane ().add (allPanel);
      this.pack ();
      //this.update ();
      this.setModal (true);
    }

    public void edit ()
    {
       GenericLoaderEdit gle = new GenericLoaderEdit (classes);
       String name = (String) listOfOptions.getSelectedValue ();
       gle.setLoaderName (name);
       gle.setProperties (classParams.getProperty (name));
       gle.setClassName (classNames.getProperty (name));
       //       gle.setNameEditable (false);
       gle.setVisible (true);
       if (gle.edited ())
	 {
	   if (gle.getLoaderName () != name)
	     {
	       int i = classes.indexOf (name);
	       delete ();
	       classes.add (i, gle.getLoaderName ());
	     }    
	   classNames.setProperty (gle.getLoaderName (), 
				   gle.getClassName ());
	   classParams.setProperty (gle.getLoaderName (), 
				    gle.getProperties ());
	   listUpdate ();
	 }
    }

    public void add ()
    {
      GenericLoaderEdit gle = new GenericLoaderEdit (classes);
      gle.setVisible (true);
      if (gle.edited ())
	{
	  String name = gle.getLoaderName ();
	  classes.add (name);
	  classNames.setProperty (name, gle.getClassName ());
	  classParams.setProperty (name, gle.getProperties ());
	}
      listUpdate ();
    }
    
    private void listUpdate ()
    {
      listOfOptions.setListData (new Vector (classes));
    }
    
    
    public void delete ()
    {
      String name = (String) listOfOptions.getSelectedValue ();
      classParams.remove (name);
      classNames.remove (name);
      classes.remove (name);
      listUpdate ();
    }

    public String getSelection ()
    {
      if (chose == false)
	return null;
      return (String) listOfOptions.getSelectedValue ();
    }


    class ListListener implements ListSelectionListener
    {
      GenericSelectionList owner;
      
      ListListener (GenericSelectionList _owner)
      {
	owner = _owner;
      }
      
      public void valueChanged(ListSelectionEvent e)
      {
	JList source = (JList) e.getSource ();
	if (!source.isSelectionEmpty ())
	  {
	    owner.localActionMap.get ("Delete").setEnabled (true);
	    owner.localActionMap.get ("Edit").setEnabled (true);
	    owner.localActionMap.get ("Ok").setEnabled (true);
	  }
	else
	  {
	    owner.localActionMap.get ("Delete").setEnabled (false);
	    owner.localActionMap.get ("Edit").setEnabled (false);
	    owner.localActionMap.get ("Ok").setEnabled (false);
	  }
      }
    }
    
    class okActionListener extends AbstractAction
    {
      GenericSelectionList owner;

      okActionListener (GenericSelectionList _owner)
      {
	super ("Ok");
	owner = _owner;
	this.setEnabled (false);
	this.putValue (Action.SHORT_DESCRIPTION, 
		       "Load model from the source selected.");
      }
      
      public void actionPerformed (ActionEvent e)
      {
	owner.chose = closeModel ();
	owner.setVisible (false);
      }
    }

    class cancelActionListener extends AbstractAction
    {
      GenericSelectionList owner;

      cancelActionListener (GenericSelectionList _owner)
      {
	super ("Cancel");
	owner = _owner;
	this.putValue (Action.SHORT_DESCRIPTION, "Cancel selection of type.");
      }
      
      public void actionPerformed (ActionEvent e)
      {
	owner.chose = false;
	owner.setVisible (false);
      }
    }

    class editAction extends AbstractAction
    {
      GenericSelectionList owner;

      editAction (GenericSelectionList _owner)
      {
	super ("Edit");
	owner = _owner;
	this.setEnabled (false);
	this.putValue (Action.SHORT_DESCRIPTION, "Edit the settings for this loader");
      }
    
      public void actionPerformed (ActionEvent e)
      {
	owner.edit ();
      }
    }
      
   
    class deleteAction extends AbstractAction
    {
      GenericSelectionList owner;

      deleteAction (GenericSelectionList _owner)
      {
	super ("Delete");
	owner = _owner;
	this.setEnabled (false);
      }
	
      public void actionPerformed (ActionEvent e)
      {
	owner.delete ();
      }
      
    }

    class addAction extends AbstractAction
    {
      GenericSelectionList owner;
      
      addAction (GenericSelectionList _owner)
      {
	super ("add loader");
	owner = _owner;
      }
      
      public void actionPerformed (ActionEvent e)
      {
	owner.add ();
      }
    }
    
  }


  /**
   ** load a generic model
   **/
  private void setupGenericModel ()
  {
    //GenericModelEdit gme = new GenericModelEdit ();
    //gme.setVisible (true);
    //String name = gme.getName ();
    //loaderClassNames.setProperty (name, gme.getClassName ());
    //loaderProperties.setProperty (name, gme.getProperties ());
  }

  private void editGenericModel (String name)
  {
//     GenericModelEdit gme = new GenericModelEdit ();
//     gme.setInitialName (name);
//     gme.setInitialProperties (loaderProperties.getProperty (name));
//     gme.setInitialClassName (loaderClassNames.getProperty (name));
//     gme.setNameEditable (false);
//     gme.setVisible (true);
//     loaderClassNames.setProperty (name, gme.getClassName ());
//     loaderProperties.setProperty (name, gme.getProperties ());
  }

  class GenericLoaderEdit extends JDialog
  {
    ActionMap localActionMap = new ActionMap ();
    JTextField nameField = new JTextField ();
    JTextField classNameField = new JTextField ();
    JTextField propertyField = new JTextField ();
    boolean userEdited;
    java.util.List takenNames;
    String originalName = "";

    GenericLoaderEdit (java.util.List _takenNames)
    {
      super ((Frame) null, "Loader details", true);
      
      takenNames = _takenNames;

      JPanel allPanel = new JPanel ();
      JPanel fieldPanel = new JPanel ();
      JPanel buttonPanel = new JPanel ();
      
      allPanel.setLayout (new BoxLayout (allPanel, BoxLayout.Y_AXIS));
      fieldPanel.setLayout (new BoxLayout (fieldPanel, BoxLayout.Y_AXIS));
      buttonPanel.setLayout (new BoxLayout (buttonPanel, BoxLayout.X_AXIS));

      JButton ok = new JButton (new okAction (this));
      localActionMap.put (ok.getAction ().getValue (Action.NAME),
			  ok.getAction ());

      JButton cancel = new JButton (new cancelAction (this));
      localActionMap.put (cancel.getAction ().getValue (Action.NAME),
			  cancel.getAction ());



      DocumentListener t = new TextChanged (this);
      
      nameField.getDocument ().addDocumentListener (t);
      classNameField.getDocument ().addDocumentListener (t);
      propertyField.getDocument ().addDocumentListener (t);

      fieldPanel.add (new JLabel ("Loader name:"));
      fieldPanel.add (nameField);
      fieldPanel.add (new JLabel ("Java class name:"));
      fieldPanel.add (classNameField);
      fieldPanel.add (new JLabel ("Parameters:"));
      fieldPanel.add (propertyField);
      
      buttonPanel.add (StandardFiller.makeHstrut ());
      buttonPanel.add (cancel);
      buttonPanel.add (ok);
      
      allPanel.add (fieldPanel);
      allPanel.add (buttonPanel);

      allPanel.setBorder (StandardFiller.makeWideEmptyBorder ());

      this.getContentPane ().add (allPanel);

      this.pack ();
      this.setModal (true);
    }

    public boolean edited ()
    {
      return userEdited;
    }
    

    public GenericLoaderEdit setLoaderName (String n)
    {
      nameField.setText (n);
      originalName = n;
      return this;
    }
    
    public GenericLoaderEdit setProperties (String n)
    {
      propertyField.setText (n);
      return this;
    }
    
    public GenericLoaderEdit setClassName (String n)
    {
      classNameField.setText (n);
      return this;
    }

    public void setNameEditable (boolean b)
    {
      nameField.setEditable (b);
    }

    public String getLoaderName ()
    {
      return nameField.getText ();
    }

    public String getClassName ()
    {
      return classNameField.getText ();
    }
    
    public String getProperties ()
    {
      return propertyField.getText ();
    }

    class okAction extends AbstractAction
    {
      GenericLoaderEdit owner;
      
      okAction (GenericLoaderEdit _owner)
      {
	super ("Ok");
	owner = _owner;
	this.setEnabled (false);
      }
      
      public void actionPerformed (ActionEvent e)
      {
	if (owner.getLoaderName ().equals (""))
	  {
	    JOptionPane.showMessageDialog (null, "Sorry, you did not provide a loader name. \n \n Unlike Polyphemus, we require that a name be specified.", 
					   "Loader name required.",
					  JOptionPane.ERROR_MESSAGE);
	    return;
	  }
	
	if (!owner.getLoaderName ().equals (originalName))
	  {
	    if (takenNames.contains (owner.getLoaderName ()))
	      {
		JOptionPane.showMessageDialog (null, "Unfortunately, this loader name, \"" + StringUtil.maxLength (owner.getLoaderName (), 15, "...") + ",\" is already taken. \n Please choose another.", "Name conflict.", JOptionPane.ERROR_MESSAGE);
		return;
	      }
	  }
	
	owner.userEdited = true;
	owner.setVisible (false);
      }
    }
    
    class cancelAction extends AbstractAction
    {
      GenericLoaderEdit owner;
      
      cancelAction (GenericLoaderEdit _owner)
      {
	super ("Cancel");
	owner = _owner;
	this.setEnabled (true);
      }
      
      public void actionPerformed (ActionEvent e)
      {
	owner.userEdited = false;
	owner.setVisible (false);
      }
    } 

    class TextChanged implements DocumentListener
    {
      GenericLoaderEdit owner;

      public void changedUpdate(DocumentEvent e) 
      {
	owner.localActionMap.get ("Ok").setEnabled (true);
      }

      public void insertUpdate(DocumentEvent e) 
      {
	changedUpdate (e);
      }

      public void removeUpdate(DocumentEvent e) 
      {
	changedUpdate (e);
      }
      
      TextChanged (GenericLoaderEdit _owner)
      {
	owner = _owner;
      }


    }
    
  }

  private void openModel ()
  {
    OpenModelPane d = new OpenModelPane (actionMap);
    d.setLocation ((int)((this.getBounds ().getWidth ()-d.getSize ().getWidth ())/2+this.getBounds ().getX ()), (int) ((this.getSize ().getHeight ()-d.getBounds ().getHeight ())/2 + this.getBounds ().getY ()));
    d.setVisible (true);
  }
  
  class OpenModelPane extends JDialog
  {
    OpenModelPane (ActionMap actionMap)
    {
      super ((Frame)null, "Choose Model Source", true);
      JPanel optionPanel = new JPanel ();
      optionPanel.setLayout (new BoxLayout (optionPanel, BoxLayout.Y_AXIS));
      JLabel header = new JLabel ("Obtain model from...");
      optionPanel.add (header);
      // This is the most ignominiously ridiculous wrangling in which
      // with an api to engange I have ever been reduced!
      JRadioButton xmlFile = new JRadioButton 
	("XKripke structure (XML)");
      xmlFile.setActionCommand ("Open XML model...");
      JRadioButton bddFile = new JRadioButton
	("BDD file");
      bddFile.setActionCommand ("Open BDD model...");
      JRadioButton generic = new JRadioButton
	("Custom source");
      generic.setActionCommand ("Open a generic model...");

      final ButtonGroup option = new ButtonGroup ();
      option.add (xmlFile);
      option.add (bddFile);
      option.add (generic);
      optionPanel.add (xmlFile);
      optionPanel.add (bddFile);
      optionPanel.add (generic);
      
      // As in "Kilroy was here, ok?" -- the original graffiti
      JButton kilroy = new JButton ("Acquire");
      final ActionMap am = actionMap;
      // XXX Stupid hack.
      final OpenModelPane thisthis = this;
      kilroy.addActionListener 
	(new ActionListener ()
	  {
	    public void actionPerformed (ActionEvent e)
	    {
	      am.get(option.getSelection ().getActionCommand ())
		.actionPerformed (e);
	      thisthis.setVisible (false);
	    }
	  });
      JButton wasHere = new JButton ("Cancel");
      wasHere.addActionListener 
	(new ActionListener ()
	  {
	    public void actionPerformed (ActionEvent e)
	    {
	      thisthis.setVisible (false);
	    }
	  });
      JPanel buttonPanel = new JPanel ();
      buttonPanel.setLayout (new BoxLayout (buttonPanel, BoxLayout.X_AXIS));
      buttonPanel.add (wasHere);
      buttonPanel.add (kilroy);
      JPanel allPanel = new JPanel ();
      allPanel.setLayout (new BorderLayout ());
      ((BorderLayout) allPanel
       .getLayout ()).setHgap (StandardFiller.getSize ());
      ((BorderLayout) allPanel
       .getLayout ()).setVgap (StandardFiller.getSize ());  
      allPanel.add (optionPanel, BorderLayout.CENTER);
      allPanel.add (buttonPanel, BorderLayout.SOUTH);
      allPanel.setBorder (StandardFiller.makeWideEmptyBorder ());
      this.getContentPane ().add (allPanel); 
      this.pack ();
      this.setModal (true);
    }
  }
  

  
  /**
   ** Finish loading a model once the files and options have been chosen.
   **/
  private void finishLoadingModel ()
  {
    // -- try and create a new MC Manager
    MCManager newMCManager = new MCManager (options);

    // make sure that the model has a name
    if (options.get ("model name") == null)
      options.put ("model name", "");

    // update all model-related data
    setTitle (GUINAME + ": " + getModelName ());

    // -- load new model and go ahead
    mcman = newMCManager;
    
    // -- create a macro manager
    macroManager = mcman.getCTLMacroManager ();
    
    output.append ("Model loaded. Have fun!\n");
    ctlin.clearCTLHistory ();
    ctlin.updateVariables (getVars (mcman.getPropOrder ()));
    // now the CTL input may be enabled
    GUIutil.setEnabled (ctlin, true);
    // load macros
    optionspanel.setCTLMacroManager (macroManager);
    optionspanel.setEnableTabs (true);

    // update the model info frame if it is showing
    if (modelInfo != null) modelInfo.update ();
    // redisplay the lattice information frame, if needed
    if (lattice != null) lattice.update (mcman.getMC ().getLattice ());
        
    // enable various actions
    enableAction ("Lattice info", true);
    enableAction ("Save fairness conditions", true);
    enableAction ("Load fairness conditions", true);
    enableAction ("Save CTL history", true);
    enableAction ("Load CTL history", true);
  }


  // -- closes currently opened model and saves (or asks to save)
  // -- its current state

  // returns true if the model was closed and false if the operation
  // was cancelled
  public boolean closeModel ()
  {
    // -- currently we don't do anything
    output.append ("Closing a model\n");
    
    //default title and icon
    if (ctlin.hasHistory ())
      {
	//	Object[] options = {"Yes", "No", "Cancel"};
	int answer = JOptionPane.showConfirmDialog 
	  (this,
	   "Would you like to save CTL history?",
	   "Closing the model",
	   JOptionPane.YES_NO_CANCEL_OPTION,
	   JOptionPane.QUESTION_MESSAGE);
	if (answer == JOptionPane.YES_OPTION) // save the CTL history
	  saveCTLHistory ();
	else if (answer == JOptionPane.CANCEL_OPTION)
	  return false;
      }
    
    GUIutil.setEnabled (ctlin, false); // disable ctl input
    ctlin.setCounterExampleEnabled (false); // disable counter-example
    // hide options
    optionspanel.setEnableTabs (false);
    optionspanel.clearFairness ();
    optionspanel.clearMacros ();
    return true;
  }
  
  /**
   ** Checks the CTL formula and outputs the result.
   **/
  private void checkCTL (CTLNode ctl)
  {
    PropOrderIf propOrder = mcman.getPropOrder ();

    output.append ("Processing : " + ctl + "\n");

    // -- expand the macroes -- called twice for 2 pass expansion
    ctl = macroManager.expandMacro (ctl);
    ctl = macroManager.expandMacro (ctl);

    // -- run our model checker on the property
    // -- record the time that it takes
    StopWatch stopwatch = new StopWatch ();
    MvSet mvResult = mcman.getMC ().doCTL (ctl);
    stopwatch.stop ();

    ctlin.addCTLToHistory (ctl.toString ());

    // -- print the result
    outputResult (mvResult, stopwatch);

    // write to daVinci file
    //    mvResult.writeToDaVinci ("/tmp/result.daVinci", propOrder);

    // if the counter-example is enabled - show it
    if (ctlin.isCounterExampleEnabled ())
      {
	showCounterExample (ctl);
      }
  }

  /**
   ** Converts MvSet result into appropriate format and outputs it.
   **/
  private void outputResult (MvSet result, StopWatch stopwatch)
  {
    output.append (mcman.resultToString (result));
    output.append ("Running time: " +
		   stopwatch.getTimeElapsed ()/1000.0 + "s\n");
  }

  /**
   ** Shows the counter example for the ctl formula.
   **/
  private void showCounterExample (CTLNode ctl)
  {
    // create a counter-example manager
    KEGManager kegman = new KEGManager (options, mcman);

    DGraph KGraph = kegman.doCtl (ctl);

    // setup a unique filename    
    final String filename =
      "/tmp/keg" + System.currentTimeMillis () + ".daVinci";
    
    // choose colourmap.
    Map colourMap = edu.toronto.cs.prettygraph.PrettyConstants.singletonMap;
    
    kegman.saveKegToFile (KGraph, filename, colourMap);

    try {
      Thread t = new Thread 
	(new Runnable ()
	 {
	   public void run ()
	   {
	     try {
	       Process p = Runtime.getRuntime ().exec ("daVinci " + filename);
	       p.waitFor ();
	       File f = new File (filename);
	       f.delete ();
	     } catch (Exception e) { /* do nothing */ }
	   }
	 });
      t.start ();
    } catch (Exception e) { /* do nothing */ }
 
  }

  /**
   ** Updates al sorts of model-related data.
   **/
  private String getModelName ()
  {
    return options.get ("model name").toString ();
  }

  /**
   ** Load fairness conditions from a file.
   **/
  private void loadFairness ()
  {
    Iterator it = loadFromFile ("Load fairness");
    if (it != null)
      {
	java.util.List fair = new LinkedList ();
	for (; it.hasNext ();)
	  fair.add (BinaryTreeToCTLConverter.convertToCTL
		    (it.next ().toString ()));
	optionspanel.setFairness (fair);
      }
  }

  /**
   ** Save the Fairness conditions.
   **/
  private void saveFairness ()
  {
    saveToFile (optionspanel.getFairness ().iterator (), "Save fairness");
  }

  /**
   ** Load some CTL history.
   **/
  private void loadCTLHistory ()
  {
    Iterator it = loadFromFile ("Load CTL history");
    if (it != null)
      {
	java.util.List ctlhistory = new LinkedList ();
	for (; it.hasNext ();)
	  ctlhistory.add (it.next ().toString ());
	ctlin.setCTLHistory (ctlhistory);
      }
  }

  /**
   ** Save the current CTL history.
   **/
  private void saveCTLHistory ()
  {
    saveToFile (ctlin.getCTLHistory ().iterator (),
		"Save CTL history");
  }

  /**
   ** Load data from file.
   ** @param title -- title of the load dialog.
   ** @return -- iterator over the data that was loaded; or null.
   **/
  private Iterator loadFromFile (String title)
  {
    // get the file
    filechooser.setDialogTitle (title);
    int result = filechooser.showOpenDialog (this);
    if (result == JFileChooser.APPROVE_OPTION)
      {
	// open up the file
	try
	  {
	    BufferedReader reader = new BufferedReader 
	      (new FileReader (filechooser.getSelectedFile ()));

	    // get the data
	    java.util.List data = new LinkedList ();
	    for (String s = reader.readLine (); s != null;
		 s = reader.readLine ())
	      data.add (s);

	    return data.iterator ();
	  }
	catch (Exception e)
	  { System.out.println ("'" + title + "' error: " + e); }
      }
    return null;
  }

  /**
   ** Save data to file.
   ** @param it -- iterator over the data that needs to be saved.
   ** @param title -- title of the save dialog.
   **/
  private void saveToFile (Iterator it, String title)
  {
    // get the file
    filechooser.setDialogTitle (title);
    int result = filechooser.showSaveDialog (this);

    if (result == JFileChooser.APPROVE_OPTION)
      {
	// open up the file
	try
	  {
	    BufferedWriter writer = new BufferedWriter 
	      (new FileWriter (filechooser.getSelectedFile ()));

	    // dump data into file
	    for (; it.hasNext ();)
	      {
		String s = it.next ().toString ();
		writer.write (s, 0, s.length ());
		writer.newLine ();
	      }
	    writer.close ();
	  }
	catch (Exception e)
	  { System.out.println ("'" + title + "' error:" + e); }
      }
  }

  /**
   ** Confirm the user's intention to quit.
   **/
  private void quit ()
  {
    if (mcman != null)
      {
	if (!closeModel ())
	  return;
	saveLoaders ();
	System.exit (0);
      }
    int response = JOptionPane.showConfirmDialog (null, 
				   "Quit XChek?",
				   "Confirmation",
				   JOptionPane.YES_NO_OPTION);
    if (response == JOptionPane.YES_OPTION)
      {
	saveLoaders ();
	System.exit(0);
      }
  }

  private void saveLoaders ()
  {
    try
      {
	FileOutputStream loaderFile;
	// XXX Temporary hack.
	loaderFile = new FileOutputStream (LOADERFILE+".classnames");
	loaderClassNames.store (loaderFile, "Class names for loaders.");
	loaderFile.close ();
	loaderFile = new FileOutputStream (LOADERFILE+".params");
	loaderProperties.store (loaderFile, "Parameters for loaders.");
	loaderFile.close ();
      }
    catch (Exception e)
      {
	System.err.println ("Sorry mate, your loaders just bit the dust.");
      }
  }  

  private void loadLoaders ()
  {

    try
      {
	FileInputStream loaderFile;
	loaderFile = new FileInputStream (new File (LOADERFILE+".classnames"));
	loaderClassNames.load (loaderFile);
	loaderFile.close ();
	loaderFile = new FileInputStream (new File (LOADERFILE+".params"));
	loaderProperties.load (loaderFile);
	loaderFile.close ();
	java.util.List l1 = new java.util.ArrayList (loaderClassNames.keySet ());
	java.util.List l2 = new java.util.ArrayList (loaderClassNames.keySet ());
	loaderClasses = (java.util.List)((ArrayList)l1).clone ();
	Assert.assert (l1.equals (l2));
      }
    catch (Exception e)
      {
	System.err.println (e.getClass ());
	System.err.println (e.getMessage ());
	System.err.println ("Oops. Sorry. We appear to have eaten your loader settings. Starting over...");
	loaderClasses = new ArrayList ();
	loaderProperties = new Properties ();
	loaderClassNames = new Properties ();
      }
  }
  
  /**
   ** Shows some information about the currently loaded model.
   **/
  private void showModelInfo ()
  {
    if (modelInfo == null) modelInfo = new ModelInfoFrame ();

    modelInfo.setVisible (true);
    modelInfo.setState (JFrame.NORMAL);
  }

  /**
   ** Show variables used in the currently loaded model.
   **/
  private void showVariables ()
  {
    String [] vars = null;

    if (mcman != null)
      {
	PropOrderIf propOrder = mcman.getPropOrder ();
	vars = getVars (propOrder);
      }

    ctlin.showVariables (vars);
  }
  private String [] getVars (PropOrderIf p)
  {
    int [] props = p.getUnPrimedProps ();
    String [] result = new String [props.length];
    for (int i = 0; i < props.length; i++)
      result [i] = p.getPropName (props [i]);

    return result;
  }

  // display the lattice information window
  private void showLatticeDisplay ()
  {
    if (lattice == null)
      {
	lattice = new LatticeDisplay (mcman.getMC().getLattice());
	lattice.setLocation ((int)((this.getBounds ().getWidth ()-lattice.getSize ().getWidth ())/2+this.getBounds ().getX ()), (int) ((this.getSize ().getHeight ()-lattice.getBounds ().getHeight ())/2 + this.getBounds ().getY ()));
      }
    lattice.setVisible (true);
    lattice.setState (JFrame.NORMAL);
    
  }

  /**
   ** Select the particular type of xDD to be used in MvSets.
   **/
  private void selectxDD ()
  {
    ModelOptions modelOpts = new ModelOptions (this, options);
    modelOpts.setLocation ((int)((this.getBounds ().getWidth ()-modelOpts.getSize ().getWidth ())/2+this.getBounds ().getX ()), (int) ((this.getSize ().getHeight ()-modelOpts.getBounds ().getHeight ())/2 + this.getBounds ().getY ()));
    modelOpts.setVisible (true);
  }
  
  private void showHelp ()
  {
    JOptionPane.showMessageDialog (this, "No help available just yet. Sorry.",
				   "Help", JOptionPane.ERROR_MESSAGE);
  }
  
  private void showAbout ()
  {
    JOptionPane.showMessageDialog (this, "Multi Valued Model Checker " +
				   version + "\n\n" +
				   "(c) 2002 University of Toronto",
				   "About XChek",
				   JOptionPane.INFORMATION_MESSAGE);
  }

//      String choices [] = {
//        "JMvSet",
//        "MDDMvSet",
//        "JBDDMvSet",
//        "CUBDDMvSet",
//        "CUADDMvSet",
//        "JADDMvSet",
//        "CUBiADDMvSet"};

//      // show the dialog
//      Object ddtype = JOptionPane.showInputDialog (this,
//  						 "Select the MvSet type:",
//  						 "Type of MvSet to use",
//  						 JOptionPane.PLAIN_MESSAGE,
//  						 null,
//  						 choices, "MDDMvSet");
//      if (ddtype != null) options.put ("mvset", ddtype);
//      else options.put ("mvset", "mdd"); // default
//  }


  // enable given action
  private void enableAction (String actionName, boolean enable)
  {  
    Action action = (Action) actionMap.get (actionName);
    action.setEnabled (enable);
    actionMap.put (actionName, action);
  }
  

  public final URL openXMLIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/Open24.gif");
  public final URL openBDDIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/Open24.gif");
  public final URL loadFairnessIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/Import24.gif");
  public final URL saveFairnessIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/Save24.gif");
  public final URL loadCTLIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/Import24.gif");
  public final URL saveCTLIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/Save24.gif");
  public final URL quitIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/Stop24.gif");
  public final URL modelInfoIcon = 
    getClass().getResource ("/toolbarButtonGraphics/general/About24.gif");
  public final URL latticeIcon =
    getClass().getResource ("/play/lattice.gif");
  public final URL varIcon =
    getClass().getResource ("/toolbarButtonGraphics/general/History24.gif");

  class ModelInfoFrame extends JFrame
  {
    JTextField model, mvset;
    JLabel modellabel, typelabel, latticelabel, ddtext, nolabel;
    JButton latbutton;

    ModelInfoFrame ()
    {
      JPanel panel = new JPanel ();
      panel.setLayout (new BoxLayout (panel, BoxLayout.Y_AXIS));
    
      modellabel = new JLabel ("Model: ");
      panel.add (modellabel);
      model = new JTextField ();
      model.setEditable (false);
      panel.add (model);
      panel.add (StandardFiller.makeVstrut ());
      typelabel = new JLabel ("DD type: ");
      panel.add (typelabel);
      mvset = new JTextField ();
      mvset.setEditable (false);
      panel.add (mvset);
      ddtext = new JLabel ();
      panel.add (ddtext);
      panel.add (StandardFiller.makeVstrut ());
      latticelabel = new JLabel ();
      panel.add (latticelabel);

      latbutton = new JButton ("View lattice");
      panel.add (latbutton);
      latbutton.addActionListener
	(new ActionListener ()
	 {
	   public void actionPerformed (ActionEvent e)
	     { showLatticeDisplay (); }
	 });

      if (mcman != null)
	{
	  update ();
	}
      else
	{
	  model.setVisible (false);
	  mvset.setVisible (false);
	  latticelabel.setVisible (false);
	  ddtext.setVisible (false);	  
	  latbutton.setVisible (false);
	  modellabel.setVisible (false);
	  typelabel.setVisible (false);
	  nolabel = new JLabel ("No model loaded");
	}

      if (nolabel != null)
	panel.add (nolabel);

      // we need a window and an OK button on it
      JButton ok = new JButton ("OK");
      final JFrame self = this; // needed for the action listener
      ok.addActionListener
	(new ActionListener ()
	 {
	   public void actionPerformed (ActionEvent e)
	     { self.setVisible (false); }
	 });

      // make the layout a bit nicer
      panel.setBorder (StandardFiller.makeWideEmptyBorder ());
      GUIutil.alignAllX (panel.getComponents (), Component.LEFT_ALIGNMENT);
      JPanel button = new JPanel ();
      button.setBorder (StandardFiller.makeWideEmptyBorder ());
      button.add (ok);

      // arrange the window and display it
      this.setTitle ("Model info");
      this.getContentPane ().add (panel, BorderLayout.CENTER);
      this.getContentPane ().add (button, BorderLayout.SOUTH);
      this.pack ();
    }

    void update ()
    {      
      if (nolabel != null)
	nolabel.setVisible (false);
      model.setVisible (true);
      mvset.setVisible (true);
      latticelabel.setVisible (true);
      ddtext.setVisible (true);	  
      latbutton.setVisible (true);
      modellabel.setVisible (true);
      typelabel.setVisible (true);

      model.setText (getModelName ());
      ddtext.setText ("# of DD nodes: " +
		      mcman.getMC ().getMvTransRelation ().size ());
      String mvsettype = options.get ("mvset").toString ();
      String ddname = "";

    // select MvSet type
    if (mvsettype.equals ("mdd") || mvsettype.equals ("MDDMvSet"))
      // -- MDD based
	ddname = "MDD";
    else if (mvsettype.equals ("join") || mvsettype.equals ("JMvSet"))
      // -- Join Irreducible based
	ddname = "JDD";
    else if (mvsettype.equals ("jbdd") || mvsettype.equals ("JBDDMvSet"))
      // -- JBDD implementation
	ddname = "JBDD";
    else if (mvsettype.equals ("cubdd") || mvsettype.equals ("CUBDDMvSet"))
	ddname = "CUBDD";
    else if (mvsettype.equals ("cuadd") || mvsettype.equals ("CUADDMvSet"))
	ddname = "CUADD";
    else if (mvsettype.equals ("jadd") || mvsettype.equals ("JADDMvSet"))
	ddname = "JADD";
    else if (mvsettype.equals ("cubiadd") || mvsettype.equals("CUBiADDMvSet"))
	ddname = "CUBiADD";

      mvset.setText (ddname);
      latticelabel.setText ("Using a lattice of size: " + 
			    mcman.getMC ().getLattice ().size ());
    }
  }

}
