package edu.toronto.cs.keg;

import java.util.*;
import java.io.*;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.heuristics.*;
import edu.toronto.cs.modelchecker.*;

/**
 ** This is the CTL simplifier.
 **/
public class KECTLSimplifier implements KEGExpander
{
  protected HeuristicAnalyser heuristics;
  
  /**
   ** Get the heuristics.
   ** @return value of heuristics.
   **/
  public HeuristicAnalyser getHeuristics() {
    return heuristics;
  }
  
  /**
   ** Set the heuristics.
   ** @param h -- heuristics to be used.
   **/
  public void setHeuristics(HeuristicAnalyser h) {
    heuristics = h;
  }

  /**
   ** Constructs the simplifier without heuristics.
   **/
  public KECTLSimplifier ()
  {
    heuristics = new NoHeuristics ();
  }

  /**
   ** Constructs the simplifier with specifed heuristics.
   **/
  public KECTLSimplifier (HeuristicAnalyser h)
  {
    heuristics = h;
  }

  /**
   ** Attempts to expand the CTL formula associated with the state in
   ** order to provide a counter-example.
   **
   ** @param state -- a node in the state graph that contains the state
   ** to be considered.
   **
   ** @param formula -- a node in the CTL graph that contains the
   ** formula to be considered.
   **
   ** @param mc -- the model checker to be used.
   **
   ** @return -- true if the expansion was successful; false otherwise.
   **/
  public boolean expand (DGNode state, DGNode formula, MvSetModelChecker mc)
  {
    KEGCtlNode ctlnode = (KEGCtlNode) formula;
    DGraph ctlgraph = ctlnode.getGraph ();

    MvState s0 = ((KEGStateNode) state).getState ();

    Set newctl; // -- the new CTL formulas

    /*System.out.println ("CTLExpand with lattice:");
    PrintWriter out = new PrintWriter (System.out);    
    mc.getLattice ().dumpValueMap (out);
    out.flush ();*/

    if (ctlnode.getCTL () instanceof CTLEqualsNode)
      // -- the formula has a value
      {
	// -- break the CTL into formula + value
	CTLNode ctl = ctlnode.getCTL ().getLeft ();
	CTLLatticeNode value = (CTLLatticeNode) ctlnode.getCTL ().getRight ();

	// -- make sure that the formula is correct
	if (mc.checkCTL (ctl).evaluate (s0) 
	    == value.getLatIdx ())
	  { // formula has a correct value
	    CTLEXEUEGSimplifier ctlsimplifier =
	      new CTLEXEUEGSimplifier ();
	    CTLNode simple = ctlsimplifier.simplify 
	      (ctl, mc.getLattice ());
	    if (!simple.equals (ctl)) // successfully simplified
	      {
		newctl = new HashSet ();
		newctl.add (simple.eq (value));
	      }
	    else if (ctl instanceof CTLNegNode)
	      newctl = explainNegation ((CTLNegNode) ctl, value, mc);
	    else if (ctl instanceof CTLAndNode)
	      newctl = explainAnd ((CTLAndNode) ctl, value, s0, mc);
	    else if (ctl instanceof CTLOrNode)
	      newctl = explainOr ((CTLOrNode) ctl, value, s0, mc);
	    else if (ctl instanceof CTLEUNode)
	      newctl = explainEU ((CTLEUNode) ctl, value, s0, mc);
	    else if (ctl instanceof CTLEGNode)
	      newctl = explainEG ((CTLEGNode) ctl, value, s0, mc);
	    else if (ctl instanceof CTLEqualsNode)
	      newctl = explainEquals (ctl, s0, mc);
	    else 
	      return false;
	  }
	else // the formula had an incorrect value....
	  {
	    newctl = new HashSet ();
	    newctl.add (ctl.eq (CTLFactory.createCTLLatticeNode 
				(mc.checkCTL (ctl).evaluate (s0),
				 mc.getLattice ())));
	  }
      }
    else // -- need to give the value of the formula
      {	
	newctl = new HashSet ();
	newctl.add (slapOnTheValue (ctlnode.getCTL (), s0, mc));
      }

    // -- append all those CTL formulas to the ctlnode....
    //KEGCtlNodeFactory factory = (KEGCtlNodeFactory) ctlnode.getFactory ();
    Iterator i = newctl.iterator ();
    while (i.hasNext ())
      {
	KegUtils.addCtl (state, ctlnode, (CTLNode) i.next ());
      }

    // -- if it got here it was successfully simplified
    return true;
  }

  CTLNode slapOnTheValue (CTLNode formula, MvState state,
			  MvSetModelChecker mc)
  {
    MvSet result = mc.checkCTL (formula);
    FiniteLattice lat = mc.getLattice ();

    return formula.eq (CTLFactory.createCTLLatticeNode 
		       (result.evaluate (state), lat));
  }

  // -- all the newctl = explanations

  Set explainNegation (CTLNegNode formula, CTLLatticeNode value,
		       MvSetModelChecker mc)
  {
    Set ctl = new HashSet ();

    // -- get the negation of the value of the formula
    
    FiniteLattice lat = value.getLattice ();
    int neg = lat.unOp (value.getLatIdx (), FiniteLattice.NEG);
    CTLLatticeNode negvalue =
      CTLFactory.createCTLLatticeNode (neg, lat);
    ctl.add (formula.getRight ().eq (negvalue) );

    return ctl;
  }

  Set explainAnd (CTLAndNode formula, CTLLatticeNode value, MvState state,
		  MvSetModelChecker mc)
  {
    Set ctl = new HashSet ();

    // -- fing the left and right operands of AND
    CTLNode left = formula.getLeft ();
    CTLNode right = formula.getRight ();

    // -- find the value of hte left and right operands
    int valueleft = mc.checkCTL (left).evaluate (state);
    int valueright = mc.checkCTL (right).evaluate (state);

    // get the lattice
    FiniteLattice lattice = mc.getLattice ();

    ctl.add (left.eq (CTLFactory.createCTLLatticeNode
			  (valueleft, lattice)));
    ctl.add (right.eq (CTLFactory.createCTLLatticeNode 
			   (valueright, lattice)));
    ctl = heuristics.getInterestingAnd (ctl);

    return ctl;
  }

  Set explainOr (CTLOrNode formula, CTLLatticeNode value, MvState state,
		 MvSetModelChecker mc)
  {
    Set ctl = new HashSet ();

    // -- fing the left and right operands of AND
    CTLNode left = formula.getLeft ();
    CTLNode right = formula.getRight ();

    // -- find the value of hte left and right operands
    int valueleft = mc.checkCTL (left).evaluate (state);
    int valueright = mc.checkCTL (right).evaluate (state);

    // get the lattice
    FiniteLattice lattice = mc.getLattice ();    

    ctl.add (left.eq (CTLFactory.createCTLLatticeNode 
		      (valueleft, lattice)));
    ctl.add (right.eq (CTLFactory.createCTLLatticeNode 
		       (valueright, lattice)));

    ctl = heuristics.getInterestingOr (ctl);

    return ctl;
  }

  Set explainEquals (CTLNode formula, MvState state,
		     MvSetModelChecker mc)
  {
    Set ctl = new HashSet ();
    
    if (!(formula.getLeft () instanceof CTLLatticeNode))
      ctl.add (slapOnTheValue (formula.getLeft (), state, mc));
    if (!(formula.getRight () instanceof CTLLatticeNode))
      ctl.add (slapOnTheValue (formula.getRight (), state, mc));

    return ctl;
  }

//    Set explainImpl (CTLImplNode formula, CTLLatticeNode value)
//    {
//      Set ctl = new HashSet ();
//      // -- use a -> b <=> !a | b
//      CTLNode a = formula.getLeft ();
//      CTLNode b = formula.getRight ();
//      ctl.add (a.neg ().or (b).eq (value));

//      return ctl;
//    }

//    Set explainIff (CTLIffNode formula, CTLLatticeNode value)
//    {
//      Set ctl = new HashSet ();
//      // -- use a <-> b <=> a -> b & a <- b
//      CTLNode a = formula.getLeft ();
//      CTLNode b = formula.getRight ();
//      ctl.add (a.implies (b).and (a.impliedBy (b)).eq (value));

//      return ctl;
//    }

//    Set explainAX (CTLAXNode formula, CTLLatticeNode value)
//    {
//      Set ctl = new HashSet ();

//      // -- use AX f <=> !EX!f
//      CTLNode f = formula.getRight ();
//      ctl.add (f.neg ().ex ().neg ().eq (value));

//      return ctl;
//    }

//    Set explainAF (CTLAFNode formula, CTLLatticeNode value,
//  		 MvSetModelChecker mc)
//    {
//      Set ctl = new HashSet ();

//      // -- use AF f <=> A[true U f]
//      CTLNode f = formula.getRight ();

//      FiniteLattice lat = mc.getLattice ();
//      CTLLatticeNode top = CTLFactory.createCTLLatticeNode (lat.idtimesIdx (),
//  							  lat);
//      ctl.add (top.au (f).eq (value));

//      return ctl;
//    }

//    Set explainEF (CTLEFNode formula, CTLLatticeNode value,
//  		 MvSetModelChecker mc)
//    {
//      Set ctl = new HashSet ();

//      // -- use EF f <=> E[true U f]
//      CTLNode f = formula.getRight ();

//      FiniteLattice lat = mc.getLattice ();
//      CTLLatticeNode top = CTLFactory.createCTLLatticeNode (lat.idtimesIdx (),
//  							  lat);
//      ctl.add (top.eu (f).eq (value));

//      return ctl;
//    }

//    Set explainAG (CTLAGNode formula, CTLLatticeNode value)
//    {
//      Set ctl = new HashSet ();

//      // -- use AG f <=> !EF!f
//      CTLNode f = formula.getRight ();
//      ctl.add (f.neg ().ef ().neg ().eq (value));

//      return ctl;
//    }

  Set _explainEG (CTLEGNode formula, CTLLatticeNode value, MvState state,
		 MvSetModelChecker mc)
  {
    Set ctl = new HashSet ();

    // -- XXX WOO DOO! -- Need a nice way to get a factory
    MvSetFactory factory = mc.checkCTL (value).getFactory ();
    
    // -- use EG f <=> 1) f & EX E[f U state]
    //                 2) f & EX EG f
    //                 
    CTLNode f = formula.getRight ();
    CTLNode s = CTLFactory.createCTLMvSetNode (factory.createPoint 
					       (state,
						mc.getLattice ()
						.idtimesIdx ()));
    
    CTLNode case1 = f.and (f.eu (s).ex ());
    CTLNode case2 = f.and (f.eg ().ex ());
    
    if (mc.checkCTL (case1).evaluate (state) == value.getLatIdx ()) // case 1
      {
	System.out.println ("Cycle begins here");
	ctl.add (case1.eq (value));
      }
    else // case 2
      ctl.add (case2.eq (value));
    
    return ctl;
  }


  Set explainEG (CTLEGNode formula, CTLLatticeNode value, MvState state,
		 MvSetModelChecker mc)
  {
    Set ctl = new HashSet ();

    // -- XXX WOO DOO! -- Need a nice way to get a factory
    MvSetFactory factory = mc.checkCTL (value).getFactory ();
    
    // -- use EG f <=> 1) f & EX E[f U state]
    //                 2) f & EX EG f
    //                 
    // -- use (EG f)(s) <=> f & ((EX E[f U s]) \/ EX EG (f & !s))
    CTLNode f = formula.getRight ();
    CTLNode s = CTLFactory.createCTLMvSetNode (factory.createPoint 
					       (state,
						mc.getLattice ()
						.idtimesIdx ()));

    CTLNode disjunct1 = f.eu (s).ex ();
    CTLNode disjunct2 = f.and (s.neg ()).eg ().ex ();
    
    ctl.add (f.and (disjunct1.or (disjunct2)).eq (value));
    return ctl;
  }
  
//    Set explainAU (CTLAUNode formula, CTLLatticeNode value)
//    {
//      Set ctl = new HashSet ();
//      /*
//      CTLAUiNode auictl;
//      MvSet aui;
//      int lvalue = value.getLatIdx ();
//      int i = 0;
//      // find the smallest i s.t. AUi f = AU f
//      do
//        {
//  	auictl = CTLFactory.createCTLAUiNode (formula.getLeft (), i,
//  					      formula.getRight ());
//  	aui = mc.checkCTL (auictl);
//  	i++;	
//        }
//      while (aui.evaluate (state) != lvalue);
//      */

//      // -- use A[p U q] = !(E[!q U !q & !p] | EG!q)
//      CTLNode p = formula.getLeft ();
//      CTLNode q = formula.getRight ();

//      ctl.add (q.neg ().eg ().or 
//  	     (q.neg ().eu (p.neg ().and (q.neg ()))).neg ().eq (value));
    
//      return ctl;
//    }

//    Set explainAUi (CTLAUiNode formula, CTLLatticeNode value)
//    {
//      Set ctl = new HashSet ();

//      // -- use A[f Ui p] <=> p | f & AX A[f ui-1 p] & EX A[f Ui-1 p]
//      CTLNode f = formula.getLeft ();
//      CTLNode p = formula.getRight ();
//      int i = formula.getI();
//      if (i == 0)
//        ctl.add(p.eq(value));
//      else
//        {
//  	CTLAUiNode afup = f.au (i-1, p);
//  	ctl.add (p.or (f.and (afup.ax ().and (afup.ex ()))).eq (value));
//        }

//      return ctl;
//    }

  Set explainEU (CTLEUNode formula, CTLLatticeNode value, MvState state,
		 MvSetModelChecker mc)
  {
    Set ctl = new HashSet ();

    CTLEUiNode euictl;
    MvSet eui;
    int lvalue = value.getLatIdx ();
    int i = 0;
    // find the smallest i s.t. EUi f = EU f
    do
      {
	euictl = CTLFactory.createCTLEUiNode (formula.getLeft (), i,
					      formula.getRight ());
	eui = mc.checkCTL (euictl);
	i++;
      }
    while (eui.evaluate (state) != lvalue);
    
    ctl.add (euictl.eq (value));
    return ctl;
  }

//    Set explainEUi (CTLEUiNode formula, CTLLatticeNode value)
//    {
//      Set ctl = new HashSet ();

//      // -- use E[f Ui p] <=> p | f & EX E[f ui-1 p]
//      CTLNode f = formula.getLeft ();
//      CTLNode p = formula.getRight ();
//      int i = formula.getI();

//      if (i == 0)
//        ctl.add(p.eq(value));
//      else
//        {
//  	CTLEUiNode efup = f.eu (i-1, p);
//  	ctl.add (p.or (f.and (efup.ex ())).eq (value));
//        }

//      return ctl;
//    }
}




