package edu.toronto.cs.lift;

import java.util.*;

import edu.toronto.cs.smv.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.util.*;


// XXX for testing only
import edu.toronto.cs.modelchecker.AlternatePropOrder;
import edu.toronto.cs.algebra.*;



public class Lift extends SMVCombinedModule
{
  Button[] liftBtns;
  int floors;

  int topFloor;
  int botFloor = 1;
  
  static final int DOOR = 0;
  static final int DIR = 1;

  int floorBaseIdx;
  int floorWidth;
  
  public Lift (String _name, int _floors)
  {
    super (_name);
    floors = _floors;
    topFloor = floors;

    floorBaseIdx = 2;
    floorWidth = computeFloorWidth (topFloor);
    
    vars = new String [2 + floorWidth];
    
    vars [DOOR] = "door";
    vars [DIR] = "dir";
    for (int i = floorBaseIdx; i < floorBaseIdx + floorWidth; i++)
      vars [i] = "floor" + (i - floorBaseIdx);
    

    // -- do the buttons
    liftBtns = new Button [floors];
    for (int i = 1; i <= liftBtns.length; i++)
      liftBtns [i-1] = new Button (getName () + 
				    SMVModule.VAR_SEP + "liftButton" + i);
    addSubModules (Arrays.asList (liftBtns));
  }


  // -- compute number of bits required to represent v
  int computeFloorWidth (int v)
  {
    int width = 1;

    while ((v >> width) > 0) width++;
    return width;
  }
  

  public MvSet computeTrans ()
  {
    System.out.println ("Computing trans for " + getName ());


//      int[] data = ordering.getPrimedProps ();
//      for (int i = 0; i < data.length; i++)
//        ddManager.var (data [i], top, top);
	
//      data = ordering.getUnPrimedProps ();
//      for (int i = 0; i < data.length; i++)
//        ddManager.var (data [i], top, top);

    MvSet result = computeDir ().and  
      (computeDoor ().and 
       (computeButtons ().and 
	(computeFloor ())));    
    return result;
  }
  
  MvSet computeButtons ()
  {
    MvSet result = ddManager.createConstant (lattice.idtimesIdx ());
    
    MvSet doorOpen = cur (DOOR, top);
    for (int i = 0; i < liftBtns.length; i++)
      {
	liftBtns [i].setReset (computeFloorEquals (i+1).and (doorOpen));
	result = result.and (liftBtns [i].computeTrans ());
	if (i == 0 && result instanceof MDDMvSet)
	  ((MDDMvSet)result).writeToDaVinci ("button.daVinci", ordering);
      }

    return result;
  }

  
  MvSet computeDir ()
  {
    /*
    ** up is true, down is false **
    next(dir) := case
               idle : dir;
	       floor = topFloor : false;
	       floor = botFloor : true;
	       car_call = 0 & dir = false : true;
	       car_call = 0 & dir = true : false;
	       1  : dir;
	       esac;
    */

    // -- mvset for next(dir) = dir
    MvSet dirIsDir = ddManager.createProjection (getNextVarOrder (DIR));
    dirIsDir = dirIsDir.eq (ddManager.createProjection (getVarOrder (DIR)));

    // -- next (dir) = up
    MvSet dirIsUp = next (DIR, top);
    // -- next (dir) = down
    MvSet dirIsDown = next (DIR, bot);

    return new CaseStatement ().
      addCase (computeIdle (), dirIsDir).
      addCase (computeFloorEquals (topFloor), dirIsDown).
      addCase (computeFloorEquals (botFloor), dirIsUp).
      addCase (cur (DIR, bot).and (computeCarCallEquals (0)), 
			dirIsUp).
      addCase (cur (DIR, top).and (computeCarCallEquals (0)),
			dirIsDown).
      addDefault (dirIsDir).compute ();
  }
  
  MvSet computeFloor ()
  {
    MvSet floorIsFloor = computeFloorIsFloor ();
    MvSet floorIsFloorPlusOne = computeFloorIsFloorPlusOne ();
    MvSet floorIsFloorMinusOne = computeFloorIsFloorMinusOne ();
    
    MvSet doorIsOpen = cur (DOOR, top);
    MvSet dirIsDown = cur (DIR, bot);
    MvSet dirIsUp = cur (DIR, top);
    
    return new CaseStatement ().
      addCase (doorIsOpen, floorIsFloor).
      addCase (computeCarCallEquals (0), floorIsFloor).
      addCase (dirIsUp.and (computeFloorLess (topFloor)), 
	       floorIsFloorPlusOne).
      addCase (dirIsDown.and (computeFloorGreater (botFloor)), 
	       floorIsFloorMinusOne).
      addDefault (floorIsFloor).
      compute ();
  }

  // -- compute 'floor = v'
  MvSet computeFloorEquals (int v)
  {
    return computeNatEquals (floorBaseIdx, floorWidth, v);
  }
  
  // -- computes 'floor < v'
  MvSet computeFloorLess (int v)
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = botFloor; i < v; i++)
      result = result.or (computeNatEquals (floorBaseIdx, floorWidth, i));
    return result;
  }
  
  MvSet computeFloorGreater (int v)
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = topFloor; i > v; i--)
      result = result.or (computeNatEquals (floorBaseIdx, floorWidth, i));
    return result;
  }
  
  MvSet computeDoor ()
  {
    /**
       door := case
                floor = car_call : true;
                1 : false;
               esac
    */

    MvSet curCase = new CaseStatement ().
      addCase (computeFloorIsCarCall (), cur (DOOR, top)).
      addDefault (cur (DOOR, bot)).compute ();

    // -- we also need to compute the next case for the door, 
    // -- that is
    /* next(door) := case
         next(floor) = next (car_call) : true;
	 1 : false;
	 esac;
    **/
    // -- we acheive this by shifting the variables in the resulting mvset
    curCase = curCase.and (curCase.renameArgs (ordering.getPrimedMap ()));
    return curCase;
    
  }
  
  MvSet computeFloorIsCarCall ()
  {
    MvSet dirIsDown = cur (DIR, bot);
    MvSet dirIsUp = cur (DIR, top);
    
    
    CaseStatement case1 = new CaseStatement ();
    for (int i = topFloor; i > botFloor; i--)
      case1.addCase (liftBtns [i - 1].cur (Button.REQUEST).and
		     (computeFloorGreater (i - 1)),
		     computeFloorEquals (i));
    
    case1.addCase (liftBtns [0].cur (Button.REQUEST), 
		   computeFloorEquals (botFloor));
    case1.addDefault (ddManager.createConstant (bot));

    
    CaseStatement case2 = new CaseStatement ();
    for (int i = botFloor + 1; i <= topFloor; i++)
      case2.addCase (liftBtns [i - 2].cur (Button.REQUEST).and
		     (computeFloorLess (i)),
		     computeFloorEquals (i - 1));
    case2.addCase (liftBtns [liftBtns.length - 1].cur (Button.REQUEST),
		   computeFloorEquals (topFloor));
    case2.addDefault (ddManager.createConstant (bot));

    return dirIsDown.and (case1.compute ()).or 
      (dirIsUp.and (case2.compute ()));
  }



  MvSet computeFloorIsFloor ()
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = botFloor; i <= topFloor; i++)
      result = result.or 
	(computeNatEquals (floorBaseIdx, floorWidth, i).and
	 (computeNextNatEquals (floorBaseIdx, floorWidth, i)));
    return result;
  }
  MvSet computeFloorIsFloorPlusOne ()
  {
    MvSet result = ddManager.createConstant (bot);
    
    for (int i = botFloor; i < topFloor; i++)
      result = result.or
	(computeNatEquals (floorBaseIdx, floorWidth, i).and
	 (computeNextNatEquals (floorBaseIdx, floorWidth, i+1)));
    return result;
  }
  MvSet computeFloorIsFloorMinusOne ()
  {
    MvSet result = ddManager.createConstant (bot);
    for (int i = topFloor; i > botFloor; i--)
      result = result.or
	(computeNatEquals (floorBaseIdx, floorWidth, i).and
	 (computeNextNatEquals (floorBaseIdx, floorWidth, i-1)));
    return result;
  }
  
  
  MvSet computeCarCallEquals (int num)
  {
    if (num != 0) Assert.assert ("Can only compute 'car_call = 0'");
    
    MvSet dirIsDown = cur (DIR, bot);
    MvSet dirIsUp = cur (DIR, top);
    
    MvSet cond1;
    MvSet cond2;

    cond1 = liftBtns [0].cur (Button.REQUEST);
    
    // -- compute a disjunction of 
    // -- liftButX.REQUEST & floor > (X-1)
    for (int i = botFloor; i < topFloor; i++)
      cond1 = cond1.or 
	(computeFloorGreater (i).and
	 (liftBtns [i].cur (Button.REQUEST)));
    
    
    cond2 = liftBtns [liftBtns.length - 1].cur (Button.REQUEST);

    for (int i = botFloor + 1; i <= topFloor; i++)
      cond2 = cond2.or
	(computeFloorLess (i).and
	 (liftBtns [i-2].cur (Button.REQUEST)));
    

    return dirIsDown.and (cond1.not ()).or (dirIsUp.and (cond2.not ()));
  }
  
  
  /***
   * DEFINE
   *  idle = !liftBut1.REQUEST &
             !liftBut2.REQUEST & ... & !liftBut(topFloor).REQUEST
  */
  MvSet computeIdle ()
  {
    MvSet result = ddManager.createConstant (top);
    
    for (int i = 0; i < liftBtns.length; i++)
      result = result.and (liftBtns [i].cur (liftBtns [i].REQUEST, bot));
    return result;    
  }

  public MvState initializeInitState (MvState _initState)
  {
    MvState initState = super.initializeInitState (_initState);
    
    initState.setValue (getVarOrder (DOOR), bot);
    initState.setValue (getVarOrder (DIR), bot);
    
    // -- set 'init (floor) = 1'
    initState.setValue (getVarOrder (floorBaseIdx), top);
    for (int i = 1; i < floorWidth; i++)
      initState.setValue (getVarOrder (floorBaseIdx + i), bot);

    // XXX debug
//      initState.setValue (getVarOrder (floorBaseIdx), bot);
//      initState.setValue (getVarOrder (floorBaseIdx + 1), top);
//      initState.setValue (liftBtns [0].getVarOrder (liftBtns [0].REQUEST), top);
    return initState;
  }


  public static void main (String[] args) throws Exception
  {
    Lift lift = new Lift ("lift", 3);
    
    System.out.println ("The variables in the system are: ");
    System.out.println (Arrays.asList (lift.getVars ()));

    AlternatePropOrder propOrder = new AlternatePropOrder (lift.getVars ());
    lift.setVarOrdering (propOrder);
    System.out.println ("Order of door is " + lift.getVarOrder (DOOR));
    System.out.println ("Order of next(door) is " + 
			lift.getNextVarOrder (DOOR));

    FiniteLattice lattice = LatticeFactory.parse ("play/gxl/boolean.xml");
    
    MvSetFactory factory = 
      CUBDDMvSet.newMvSetFactory (lattice, propOrder.size ());
    
    System.out.println ("Setting factory");
    lift.setLattice (lattice);
    lift.setMvSetFactory (factory);
    System.out.println ("Computing transition relation");
    MvSet trans = lift.computeTrans ();
    MvState initState = new MvState (propOrder.size (), lattice);
    initState = lift.initializeInitState (initState);
    System.out.println ("Init state is:");
    System.out.println (initState.toString (propOrder));
    
  }

  public static SMVModule initModule (String[] args)
  {
    System.out.println ("initModule is called with args:");
    System.out.println (Arrays.asList (args));
    
    String name = args [0];
    String floorsStr = args [1];
    
    Lift lift;
    FiniteLattice lattice;
    
    try {
      System.out.println ("Intiailizing " + Lift.class.getName () + 
			  " (" + name + ", " + floorsStr + ")");
      lift = new Lift (name, Integer.parseInt (floorsStr));
      lattice = LatticeFactory.parse ("play/gxl/9val-logic.xml");
      //lattice = LatticeFactory.parse ("play/gxl/boolean.xml");
      //lattice = LatticeFactory.parse ("play/gxl/3val-logic.xml");
    }
    catch (Exception ex) {
      throw new RuntimeException ("No lattice to be found");
    }
    
    lift.setLattice (lattice);
    return lift;
  }

  /*****
   ***** Some properties of this lift to verify
   *****
   !AG (lift_door = FF -> AF (lift_door = TT))  -- checked

   AG (lift_liftButton2_request -> AF (lift_floor0 = F /\ lift_floor1 = T /\ lift_door = T))   -- checked
   
   
   EF (lift_floor0 = T /\ lift_floor1 = F /\ lift_door = F /\ lift_liftButton3_request = F /\ lift_liftButton2_request = F /\ lift_liftButton1_request = F) -- checked

  AG ( (lift_floor0 = T /\ lift_floor1 = F /\ lift_door = F /\ lift_liftButton3_request = F /\ lift_liftButton2_request = F /\ lift_liftButton1_request = F) -> EG (lift_floor0 = T /\ lift_floor1 = F /\ lift_door = F)) -- checked

  */
  
}




/****
   AG ((lift_floor0 = F /\ lift_floor1 = T /\ lift_floor2 = F /\ lift_liftButton5_request = T /\ lift_dir = T) -> A[lift_dir=T U lift_floor0 = T /\ lift_floor1 = F /\ lift_floor2 = T])  -- checked
***/
