package edu.toronto.cs.lift;

import edu.toronto.cs.smv.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.algebra.*;

import edu.toronto.cs.util.*;

import java.util.Arrays;

public class MainLiftRBG extends MainLiftRAbstract
{

  boolean optimistic = true;
  FiniteLattice twoVal;
  FiniteLattice threeVal;
  
  MvSetFactory dstFactory;
  
  public MainLiftRBG (String name, int floors, boolean _optimistic)
  {
    super (name, floors);
    optimistic = _optimistic;
    try {
      twoVal = LatticeFactory.parse ("play/gxl/boolean.xml");
      threeVal = LatticeFactory.parse ("play/gxl/3val-logic.xml");
    } catch (Exception ex) {
      Assert.assert ("Hmm", ex);
    }
    setLattice (threeVal);
  }
  
  public void setMvSetFactory (MvSetFactory factory)
  {
    MvSetFactory mFactory = 
      MDDMvSet.newMvSetFactory (threeVal, ordering.size ());
    super.setMvSetFactory (mFactory);
    dstFactory = factory;
  }
  
  public FiniteLattice getLattice ()
  {
    return twoVal;
  }
  
  public MvSet computeTrans ()
  {

    System.out.println ("Computing 3-val transition relation");
    MvSet result = super.computeTrans ();
    
    Assert.assert (result instanceof MDDMvSet, "we only support MDDs");

    MDDMvSet.MDDMvSetFactory mddFactory = 
      (MDDMvSet.MDDMvSetFactory)((MDDMvSet)result).getFactory ();
    MDDNode resultNode = ((MDDMvSet)result).getMddNode ();

    MDDManager mddManager = mddFactory.getMddManager ();

    System.out.println ("Inserting bar variables");
    resultNode = insertBarVars (mddManager, resultNode);

    
    System.out.println ("Performing resolution following mode: " + 
			(optimistic ? "optimistic" : "pessimistic"));
    resultNode = 
      mddManager.reduce (resultNode, lattice.getElementIdxByKey ("M"), 
			 optimistic ? top : bot);
    
    System.out.println ("Moving the diagram to boolean");
    MDDManager dstManager = 
      ((MDDMvSet.MDDMvSetFactory)dstFactory).getMddManager ();
    int[] logicMap = new int [3];
    logicMap [threeVal.getElementIdxByKey ("T")] 
      = twoVal.getElementIdxByKey ("T");
    logicMap [threeVal.getElementIdxByKey ("M")] 
      = Algebra.NO_VALUE;
    logicMap [threeVal.getElementIdxByKey ("F")]
      = twoVal.getElementIdxByKey ("F");
    
    resultNode = mddManager.transfer (dstManager, resultNode, logicMap);
    
    return ((MDDMvSet.MDDMvSetFactory)dstFactory).createMvSet (resultNode);
    
  }

  MDDNode insertBarVars (MDDManager mddManager, MDDNode node)
  {
    // -- for every variable x with index i, var with index i+2 is its bar
    // -- variant
    if (node.isConstant ()) return node;
    
    int varIndex = node.getVarIndex ();
    int barIndex = varIndex + 2;
    

    
    MDDNode[] nodeChildren = new MDDNode [node.howManyChildren ()];
 

    for (int i = 0; i < nodeChildren.length; i++)
      {
	
	MDDNode barNode = 
	  mddManager.getVarEquals (barIndex, 
				   lattice.unOp (i, FiniteLattice.NEG));
	nodeChildren [i] = 
	  mddManager.Apply (barNode, 
			    insertBarVars (mddManager, node.getChild (i)),
			    FiniteLattice.MEET);
      }
    return mddManager.makeUnique (varIndex, nodeChildren);    
  }  

  // -- override to add bar variables
  public String[] getVars ()
  {
    // -- these are the real variables
    String[] vars = super.getVars ();
    
    String[] barVars = new String [vars.length << 1];
    for (int i = 0; i < barVars.length; i++)
      {
	if ((i & 1) == 0)
	  barVars [i] = vars [i >> 1];
	else
	  barVars [i] = barVars [i - 1] + "bar";
      }
    return barVars;
  }
  
  
  public MvState initializeInitState (MvState initState)
  {
    MvState mInitState = new MvState (ordering.size (), threeVal);
    
    super.initializeInitState (mInitState);
    // -- now update for bar variables
    for (int i = mInitState.getFirstVar (); 
	 i <= mInitState.getLastVar (); i++)
      {
	if (!ordering.getPropName (i).endsWith ("bar") &&
	    mInitState.isSet (i))
	  mInitState.setValue (i + 2, 
			      lattice.unOp 
			      (mInitState.getValue (i), FiniteLattice.NEG));
      }
    
    return transferToBoolean (mInitState, initState);
  }

  MvState transferToBoolean (MvState src, MvState dest)
  {
    int t = twoVal.getElementIdxByKey ("T");
    int tt = threeVal.getElementIdxByKey ("T");
    int f = twoVal.getElementIdxByKey ("F");
    int ff = threeVal.getElementIdxByKey ("F");
    
    int[] r = src.getRestrictions ();
    
    for (int i = 0; i < r.length; i++)
      {
	if (r [i] == tt)
	  dest.setValue (i, t);
	else if (r [i] == ff)
	  dest.setValue (i, f);
      }
    return dest;
  }
  

  // -- Module intialization routine to be used by the model checker
  public static SMVModule initModule (String[] args)
  {
    System.out.println ("Intializing: " + MainLiftRBG.class.getName ());
    System.out.println ("Will refer to this module as \'main\'");
    String moduleName = "main";
    System.out.println ("Expecting one argument: numer of floors");

    int floors;
    try {
      floors = Integer.parseInt (args [0]);
    } catch (Exception ex) {
      System.out.println ("Hmm no arguments, got: " + ex);
      System.out.println ("Assuming 3 floors");
      floors = 3;
    }    
    MainLiftRBG main = new MainLiftRBG (moduleName, floors, args.length == 1);
    return main;
  }
  
  
}
