package edu.toronto.cs.mdd;

import java.io.*;
import java.util.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.algebra.*;


import edu.toronto.cs.mvset.*;
import edu.toronto.cs.experiments.*;
import edu.toronto.cs.modelchecker.*;


/*** Dumps a DD into a file ***/
public class DDDumper
{

  /****
   **** File format is as follows
   **** identifier: "DD"
   **** (blockid blockdata)+
   **** Where blockid can be LATTICE_NAME -> block_data is a name of the 
                                             lattice
					     
	                     DD_DATA -> block_data is the DD children
			     have even indexes, odd correpond to
			     negation, 0 child corresponds to
			     no-child, used to identify terminals
   
  ****/

  public static final byte LATTICE_NAME = 1;
  public static final byte DD_DATA = 2;
  // -- incoming complement edge
  public static final byte DD_DATA_COMPELEMENT = 3;
  
  public static final char[] id = {'D', 'D'};

  public static final int NO_CHILD = 0;
  
  MarshaledOutputStream out;
  Algebra algebra;
  String latticeName;

  int lastWritten;
  int arrity;
  
  // -- keeps track of already written diagrams
  Map written;

  public DDDumper (OutputStream _out, String _latticeName, Algebra _algebra)
  {
    out = new MarshaledOutputStream (_out);
    algebra = _algebra;
    latticeName = _latticeName;
  }
  public DDDumper (MarshaledOutputStream _out, String _latticeName,
		   Algebra _algebra)
  {
    out = _out;
    algebra = _algebra;
    latticeName = _latticeName;
  } 
  

  public void writeDD (MDDNode node) throws IOException
  {
    lastWritten = 0;
    written = new HashMap ();

    // XXX This means we cannot write constant nodes normally!
    arrity = node.howManyChildren ();
  
    writeHeaders ();
    out.write (DD_DATA);
    out.writeShort ((short)node.howManyChildren ());
    writeDDRecur (node);
  }
  
  void writeHeaders () throws IOException
  {
    out.writeCharArrayFixed (id);
    out.writeByte (LATTICE_NAME);
    out.writeString (latticeName);
  }
  
  // -- writes a node and all its children, returning this nodes index
  int writeDDRecur (MDDNode node) throws IOException
  {
    
    // -- check if this node has been written before, if so 
    // -- just return its index
    Integer nodePtr = (Integer)written.get (node);
    if (nodePtr != null)
      return nodePtr.intValue () << 1;
    

    // -- place to store pointers to children
    int[] children = new int [arrity];



    // -- write all children and var index or value in case of constant
    if (node.isConstant ())
      {
	Arrays.fill (children, NO_CHILD);
	out.writeInt (node.getValue ());
      }
    else
      {
	// -- write all the children
	for (int i = 0; i < children.length; i++)
	  children [i] = writeDDRecur (node.getChild (i));
	out.writeInt (node.getVarIndex ());
      }
//     System.out.println ("Writing node: " +
// 			(node.isConstant () ? "C" + node.getValue () :
// 			 String.valueOf (node.getVarIndex ())));
    
    // -- write the children array
    out.writeIntArrayFixed (children);
    
    // -- increment last written counter
    lastWritten++;
    // -- add to the written table
    written.put (node, new Integer (lastWritten));
    // -- all indexes should be even
    return lastWritten << 1;
  }
  
  
  // -- test method
  public static void main (String[] args) throws Exception
  {
    
    FiniteLattice lattice = Lattices.getLatticeByName ("2Bool");
    
    MvSetFactory factory = MDDMvSet.newMvSetFactory (lattice, 4);
    
    int[] data = new int[] {0, 0, 1, 1};
    
    MvSet mvSet = factory.createPoint (data, lattice.idtimesIdx ());
    data = new int [] {1, 0, 1, 0};
    mvSet = mvSet.or (factory.createPoint (data, lattice.idtimesIdx ()));

    MDDNode node = ((MDDMvSet)mvSet).getMddNode ();

    PropOrderIf ordering = new AlternatePropOrder (new String [] {"x", "y"});
    
    node.writeToDaVinci ("/tmp/node1.daVinci", ordering);
    
    FileOutputStream out = new FileOutputStream ("/tmp/node.mdd");
    DDDumper dumper = new DDDumper (out, "2Bool", lattice);
    dumper.writeDD (node);
    out.flush ();
    out.close ();


    MDDManager manager = ((MDDMvSet.MDDMvSetFactory)factory).getMddManager ();
    
    FileInputStream in = new FileInputStream ("/tmp/node.mdd");
    DDUnDumper unDumper = new DDUnDumper (in, manager);
    MDDNode nodeRead = unDumper.readDD ();

    nodeRead.writeToDaVinci ("/tmp/node2.daVinci", ordering);

    if (nodeRead == node)
      System.out.println ("Seems to work");
    else
      System.out.println ("Failed!");

//     MDDPoint point = manager.getRandomPoint (nodeRead,
// 					    new UniformChaosIterator
// 					    (new int[]{0,1}));
    
//     System.out.println ("This point has a value: " + point.getValue ());
//     point.getNode ().writeToDaVinci ("/tmp/point.daVinci", ordering);

//     nodeRead = manager.Apply (nodeRead, point.getNode (),
// 			      FiniteLattice.JOIN);
//     nodeRead.writeToDaVinci ("/tmp/result.daVinci", ordering);
  }
  
  
}
