package edu.toronto.cs.mdd;

import java.util.*;
import java.io.BufferedInputStream;
import java.io.FileInputStream;

import edu.toronto.cs.algebra.*;

import edu.toronto.cs.util.*;
import edu.toronto.cs.mvset.MvSet;
import edu.toronto.cs.mvset.MvState;
import edu.toronto.cs.mvset.MvSetFactory;


/** The manager for multiple-valued decision diagrams, first
    described in (Srinivasan et al., <i>journal</i>, year, pages).
    This version works with any indexed semiring.
    @version 1.0
    @author B. Devereux, Department of Computer Science, University of Toronto (<tt>benet@cs.toronto.edu</tt>)
*/
public class MDDMgr implements MDDManager 
{

  static Logger log = new Logger (MDDMgr.class.getName ());


  
  // -- op code for the existAbstract operation, hopefully 100 is not used
  // -- yet :)
  static final int EXIST_ABSTRACT = 100;
  static final int RENAME_VARS = 101;
  static final int FORALL_ABSTRACT = 102;
  
  static String cacheImplName; // name of cache implementations
  static Class cacheImpl; // the cache implementation

  /** This is the static initializer. It attempts to open the
      properties file "MDDMgr.properties" and read the property
      "mddmgr.cache" from it. This is the (full) name of the
      implementation. If this can't be found, the default
      implemetation "edu.toronto.cs.MDD.MDDCacheStub" is used.
  */
  static {
	
    try {
      Properties props = new Properties();
      props.load(new BufferedInputStream
		 (new FileInputStream("MDDMgr.properties")));
      cacheImplName = props.getProperty("mddmgr.cache");
	    
    } catch (Exception e) {
      cacheImplName="edu.toronto.cs.mdd.MDDCacheStub";
    }

    try {
      log.debug ("Loading class " + cacheImplName);
      cacheImpl = Class.forName(cacheImplName);
    } catch (Exception e) {
      log.debug ("MDDMgr: initializer error: " + e);
    }
						   
  };
  int n; // number of variables
  Algebra s;
  int termvals; // number of terminal values
  int plicity; // arity of internal nodes

  MDDNode constants[]; // constant nodes
  Map uniqueTable;
  MDDNodeCache cache;

  MDDNode top;
  MDDNode bot;
  

  // XXX This assumes that all calls to MDDManager are synchronized!
  MDDNode lookupNode;
  
  boolean caching = true;
  
  int[] ignoreMap;
  /* Constructor
     @param nvars number of variables
     @param ntermvals number of terminal nodes
     @param s Semiring operations and values are drawn from
     @param greatest not sure
     @param least I forget!
  */
  public MDDMgr (int nvars, Algebra sr, int[] ignoreMap)
  {
    this (nvars, sr, sr.size (), ignoreMap);
  }
  
  public MDDMgr (int nvars, Algebra sr)
  {
    this (nvars, sr, sr.size (), null);
  }
  public MDDMgr (int nvars, Algebra sr, int arrity)
  {
    this (nvars, sr, arrity, null);
  }
  
  public MDDMgr(int nvars, Algebra sr,
		int arity, int[] _ignoreMap) {
    ignoreMap = _ignoreMap;
    n = nvars;
    termvals = sr.size();
    s = sr;
    plicity = arity;
	
    uniqueTable = new HashMap ();
    try {
      cache = (MDDNodeCache) cacheImpl.newInstance();
    }
    catch (Exception e)
      {
	e.printStackTrace ();
	// CDTW changed exit to assert
	Assert.assert(e);
      }
    
    lookupNode = new MDDNode (this);

    constants = new MDDNode [termvals];
    for (int i=0; i < termvals; i++)
      constants [i] = getConstant (i);

    // -- set top and bot constants as special
    top = getConstant (sr.idtimesIdx ());
    bot = getConstant (sr.idplusIdx ());


    //doFourValOpt (sr);
  }

  public void setIgnoreMap (int[] v)
  {
    ignoreMap = v;
  }
  
  
  boolean fourValOpt = false;
  int tt = -1;
  int tf = -1;
  int ft = -1;
  int ff = -1;
  
  int optWorked = 0;
  int optFailed = 0;
  
  void doFourValOpt (Algebra algebra)
  {
    FiniteLattice lattice;
    if (!(algebra instanceof FiniteLattice))
      return;
    lattice = (FiniteLattice)algebra;
    if (lattice.size () != 4) return;
    
    System.out.println ("Enabling fourVal optimization!");
    fourValOpt = true;
    tt = lattice.getElementIdxByKey ("TT");
    tf = lattice.getElementIdxByKey ("TF");
    ft = lattice.getElementIdxByKey ("FT");
    ff = lattice.getElementIdxByKey ("FF");
  }
  
  public void setCaching (boolean v)
  {
    caching = v;
    // -- clear the cache
    //      cache.cleanup ();
    //      uniqueTable = new Hashtable ();
  }
  
  public int getNumVars ()
  {  
    return n; 
  }

  // -- returns an algebra used by this MDDManager
  public Algebra getAlgebra ()
  {
    return s;
  }
  
  //    public String toString() {
  //      return uniqueTable.toString();
  //    }
  public String toString ()
  {
    return statsToString ();
  }
  
  int[] cacheHits = new int []{0,0,0,0,0};
  int[] opCount = new int []{0,0,0,0,0};
  
  public void resetStats ()
  {
    for (int i = 0; i < cacheHits.length; i++)
      {
	cacheHits [i] = 0;
	opCount [i] = 0;
      }
    cache.resetStats ();
  }
  
  void countCacheHit (int op)
  {
    //cacheHits [op] += 1;
  }

  void countOp (int op)
  {
    //opCount [op] += 1;
  }
  
  public String statsToString ()
  {
    String[] ops = new String[] {"and", "or", "not", "existAbstract", 
				 "forallAbstract"};

    StringBuffer sb = new StringBuffer ();
    for (int i = 0; i < cacheHits.length; i++)
      {
	sb.append (ops [i]);
	sb.append (": ");
	sb.append (cacheHits [i]);
	sb.append (" hits, ");
	sb.append ("op count: ");
	sb.append (opCount [i]);
	sb.append (" total\n");
      }
    sb.append (cache.toString ());
    sb.append ("\n");
    return sb.toString ();
    
  }
  
  

  /** Return one of the constant nodes. 
      @param val index of constant */
  public MDDNode getConstant (int val) 
  {
    
    lookupNode.setValue (val);
    MDDNode result = (MDDNode)uniqueTable.get (lookupNode);
    if (result == null)
      {
	result = new MDDNode (this, val);
	uniqueTable.put (result, result);
      }
    return result;
  }

  public MDDNode getZero() {
    return bot;
  }

  public MDDNode getOne() {
    return top;
  }

  /** Return a the one-variable node 'x_i' */
  public MDDNode getOneVar(int idx) {
    MDDNode[] constantsDst = new MDDNode [constants.length];
    System.arraycopy (constants, 0, constantsDst, 0, constants.length);
    return makeUnique(idx, constantsDst);
  }

  /** Return the one-variable node x_i = v */
  public MDDNode getVarEquals(int idx, int val) 
  {
    MDDNode kids[] = new MDDNode [plicity];
    
    MDDNode zero = getZero ();
    Arrays.fill (kids, zero);
    kids [val] = getOne ();

    return makeUnique(idx, kids);    
  } // end getVarEquals()

  /** setEqualTo */
  public MDDNode setEqualTo(MDDNode m, int val) {
    if (m.isConstant ())
      {
	if (m.equals (getOne ()))
	  return getConstant (val);
	else 
	  return getZero ();
      }
    else // m not constant 
      { 
	MDDNode kids[] = new MDDNode [plicity];
	
	for (int i = 0; i < plicity; i++)
	  kids[i] = setEqualTo (m.getChild (i), val);

	return makeUnique (m.getVarIndex (), kids);
      }
  }

  /** If the described MDDNode is already in the manager,
      return it. Otherwise create it, and then return it.
      @param varidx variable at node
      @param chdrn array of children
      @result either the existing MDDNode described by the parameters,
      or a new one.
  */
  public MDDNode makeUnique(int varidx, MDDNode[] chdrn) 
  {

    // -- apply ignore map
    if (ignoreMap != null) 
      for (int i = 0; i < ignoreMap.length; i++)
	{
	  //  	    if (chdrn [ignoreMap [i]] != getZero ())
	  //  	      Assert.assert ("Should this happen here?" + chdrn [ignoreMap [i]]);
	  chdrn [ignoreMap [i]] = getZero ();
	}
      
      
      
    // first check if it is a constant node, i.e. all children are equal
    boolean con = true;
    MDDNode fst = chdrn [0];
    for (int i = 1; i < plicity; i++)
      {
	// XXX Arie: we assume that every node comes from a unique table
	// XXX Arie: thus we can use object identity instead of equality
	if (fst != chdrn [i]) 
	  {
	    con = false;
	    break;
	  }
      }
		

    // -- if this is a constant node, bail out quickly
    if (con) 
      return fst;
    
    // -- otherwise we need to see if we have this node in our uniqueTable

    //        MDDNode temp = new MDDNode (this, varidx, chdrn);

    lookupNode.setVarIndex (varidx);
    lookupNode.setChildren (chdrn);
    MDDNode temp = lookupNode;
      
    //        if (!temp.equals (lookupNode) || temp.hashCode () != lookupNode.hashCode ())
    //  	{
    //  	  System.out.println ("temp == lookupNode: " + temp.equals (lookupNode));
    //  	  System.out.println ("temp_hash: " + temp.hashCode () + 
    //  			      " lookup_hash: " + lookupNode.hashCode ());
    //  	  Assert.assert ("We've got it");

    //  	}
    //        else
    //  	temp = lookupNode;
      
	  

    // -- use the lookup node to avoid creating too many nodes
    //        MDDNode temp = lookupNode;
    //        temp.setVarIndex (varidx);
    //        temp.setChildren (chdrn);
      

    MDDNode result = (MDDNode)uniqueTable.get (temp);
    if (result == null)
      {
	result = new MDDNode (this, varidx, chdrn);
	uniqueTable.put (result, result);
      }
    return result;
      
    //        if (uniqueTable.containsKey (temp))
    //  	return (MDDNode)uniqueTable.get (temp);
    //        else 
    //  	{
    //  	  uniqueTable.put(temp, temp);
    //  	  return temp;
    //  	} 
  } 


  //    public MDDNode build(ArbArray aa) {
  //	
  //}; // end Build

  /**  Apply a binary operation to MDDs.
       @param m1 left operand
       @param m2 right operand
       @param oper opcode for finite-algebra operation
	
  */
  public MDDNode Apply (MDDNode m1, MDDNode m2, int oper) 
  { 
    MDDNode answer = null;


    // -- first deal with the simple case when both m1 and m2
    // -- are constants
    if (m1.isConstant () && m2.isConstant ())
      {
	// -- well, if everything fails we do the work
	return getConstant (s.binOp (m1.getValue (),
				     m2.getValue (),
				     oper));
      } 

    MDDNode top = getOne ();
    MDDNode bot = getZero ();

    // -- basic optimizations for top and bot
    if (FiniteLattice.MEET == oper)
      {
	if (m1 == top) return m2;
	if (m2 == top) return m1;
	if (m1 == bot || m2 == bot) return bot;
      }
    else if (FiniteLattice.JOIN == oper)
      {
	if (m1 == top || 
	    m2 == top) return top;
	if (m1 == bot) return m2;
	if (m2 == bot) return m1;
      }

      
      
    // -- gather statistics for real operations that will take time
    if (oper == FiniteLattice.JOIN)
      countOp (1);
    else if (oper == FiniteLattice.MEET)
      countOp (0);
      
    // check cache for result.
      
    if (caching)
      answer = cache.find (m1, m2, oper);
      
    //        if (answer != null)
    //  	return answer;
      
    if (answer == null && caching && 
	(oper == FiniteLattice.JOIN || oper == FiniteLattice.MEET))
      answer = cache.find (m2, m1, oper);
      
    if (answer != null)
      {
	if (oper == FiniteLattice.JOIN)
	  countCacheHit (1);
	else if (oper == FiniteLattice.MEET)
	  countCacheHit (0);
	return answer;
      }




    // otherwise must do the computation.
    MDDNode[] kids = new MDDNode [plicity]; // temporary array to hold them

    //XXX ARIE: Missing case where m1 is non-constant and m2 IS constant
    //XXX ARIE: This works due to assumption that all constant nodes have
    //XXX ARIE: variable id set to MAX_INT

    
    // case 2: both non-constant, with the same index
    if (m1.getVarIndex () == m2.getVarIndex ())
      {
	log.debug ("APPLY: non-constant, same index");
	for (int i = 0; i < plicity; i++)
	  kids [i] = Apply (m1.getChild (i), m2.getChild (i), oper);

	answer = makeUnique (m1.getVarIndex (), kids);
      }

    // case 3: both non-constant, m1 <i m2 or m2 is constant
    else if (m1.getVarIndex () < m2.getVarIndex ())
      {
	log.debug ("APPLY: non-constant, i1<i2");
	for (int i = 0; i < plicity; i++)
	  kids [i] = Apply(m1.getChild (i), m2, oper);
	answer = makeUnique (m1.getVarIndex(), kids); 
      } // end case 3
		      
    // case 4: both non-constant, m1 >i m2, or m1 is the only constant
    else // m1 > m2
      {
	log.debug ("APPLY: non-constant, i1>i2");
	for (int i = 0; i < plicity; i++)
	  kids [i] = Apply (m1, m2.getChild (i), oper);
	answer = makeUnique (m2.getVarIndex(), kids);
      } // end case 3
	
    // -- store the result in the cache for efficiency reasons
    cache.insert(m1, m2, oper, answer);
    return answer;
  } 


  
  /**
   * <code>apply</code> applies a given function to two mdds
   *
   * @param m1 a <code>MDDNode</code> value
   * @param m2 a <code>MDDNode</code> value
   * @param function an <code>ApplyFunction</code> value
   * @return a <code>MDDNode</code> value
   */
  public MDDNode apply (MDDNode m1, MDDNode m2, ApplyFunction function) 
  { 
    MDDNode answer = null;


    // -- first deal with the simple case when both m1 and m2
    // -- are constants
    if (m1.isConstant () && m2.isConstant ())
      return getConstant (function.binApply (m1.getValue (), 
					     m2.getValue ()));

      
    // check cache for result.      
    if (caching && function.isCacheable ())
      {
	answer = cache.find (m1, m2, function.getId ());
	// -- if function is symetric, try reversing the arguments
	if (answer == null && function.isSymetric ())
	  answer = cache.find (m2, m1, function.getId ());
      }
      
      
    if (answer != null) return answer;


    // otherwise must do the computation.
    MDDNode[] kids = new MDDNode [plicity]; // temporary array to hold them

    //XXX ARIE: Missing case where m1 is non-constant and m2 IS constant
    //XXX ARIE: This works due to assumption that all constant nodes have
    //XXX ARIE: variable id set to MAX_INT

    
    // case 2: both non-constant, with the same index
    if (m1.getVarIndex () == m2.getVarIndex ())
      {
	for (int i = 0; i < plicity; i++)
	  kids [i] = apply (m1.getChild (i), m2.getChild (i), function);
	answer = makeUnique (m1.getVarIndex (), kids);
      }

    // case 3: both non-constant, m1 <i m2 or m2 is constant
    else if (m1.getVarIndex () < m2.getVarIndex ())
      {
	for (int i = 0; i < plicity; i++)
	  kids [i] = apply (m1.getChild (i), m2, function);
	answer = makeUnique (m1.getVarIndex(), kids); 
      } // end case 3
		      
    // case 4: both non-constant, m1 >i m2, or m1 is the only constant
    else // m1 > m2
      {
	for (int i = 0; i < plicity; i++)
	  kids [i] = apply (m1, m2.getChild (i), function);
	answer = makeUnique (m2.getVarIndex(), kids);
      } // end case 3
	
    // -- store the result in the cache for efficiency reasons
    if (caching && function.isCacheable ())
      cache.insert(m1, m2, function.getId (), answer);
    return answer;
  } 



  MDDNode applyOpt (int value, MDDNode node, int oper)
  {
    
    // -- value is either tf or ft
    if (oper == FiniteLattice.JOIN)
      {
	if (value == tf)
	  {
	    // (tf \/ ft) = T and (tf \/ tt) = tt
	    // -- thus if the only two reachable values are ft and tt
	    // -- just return true
	    if (!node.inImage (ff) && !node.inImage (tf)) 
	      return getOne ();
	    // -- (tf \/ ff) = tf and (tf \/ tf) = tf
	    if (!node.inImage (tt) && !node.inImage (ft))
	      return getConstant (tf);
	  }
	else // value is ft
	  {
	    // (ft \/ tf) = tt and (ft \/ tt) = tt
	    if (!node.inImage (ff) && !node.inImage (ft))
	      return getOne ();
	    // -- (ft \/ ff) = ft and (ft \/ ft) = ft
	    if (!node.inImage (tt) && !node.inImage (tf))
	      return getConstant (ft);
	  }
      }
    else if (oper == FiniteLattice.MEET)
      {
	if (value == tf)
	  {
	    // -- (tf /\ ft) == ff and (tf /\ ff) == ff
	    if (!node.inImage (tt) && !node.inImage (tf))
	      return getZero ();
	    // -- (tf /\ tt) == tf and (tf /\ tf) == tf
	    if (!node.inImage (ff) && !node.inImage (ft))
	      return getConstant (tf);
	  }
	else // value is ft
	  {
	    // -- (ft /\ tf) == ff and (ft /\ ff) == ff
	    if (!node.inImage (tt) && !node.inImage (tf))
	      return getZero ();
	    // -- (ft /\ tt) == ft and (ft /\ ft) == tf
	    if (!node.inImage (ff) && !node.inImage (tf))
	      return getConstant (ft);
	  }
      }
    return null;
  }
  
  
  /** Apply a unary operation to MDDs. */
  public MDDNode Apply (MDDNode m1, int oper)
  {
    // -- constants don't take any time
    if (m1.isConstant ())
      return getConstant (s.unOp (m1.getValue (), oper));
      
    if (oper == FiniteLattice.NEG)
      countOp (2);
      
    MDDNode answer = null;
    if (caching)
      answer = cache.find(m1, oper);
    if (answer != null)
      {
	if (oper == FiniteLattice.NEG)
	  countCacheHit (2);
	return answer;
      }
      

    // not in cache! So, we have to do the work.
    MDDNode kids[] = new MDDNode [plicity];
      
    // apply 'oper' to each child, and make new unique node
    for (int i = 0; i < plicity; i++)
      kids[i] = Apply (m1.getChild (i), oper);

    answer = makeUnique (m1.getVarIndex (), kids);

		
    cache.insert(m1, oper, answer);
    return answer;
  }

  public MDDNode applyState (MDDNode m, MvState state)
  {
    Assert.assert ("Use existAbstract instead");
    for (int i = 0; i < plicity; i++)
      {
	if (state.getValue (i) != s.NO_VALUE)
	  m = Restrict (m, i, state.getValue (i));
      }
    return m;
  }

  /** Restrict a specified variable to a certain value. */
  public MDDNode Restrict (MDDNode m, int varidx, int value)
  {
    /*** Since variables are ordered, if the top level variable in
     *** 'm' has has order higher then 'varidx', then there is no
     *** variable with order 'varidx' in the MDD and we can bail out
     *** right away
     ***/
    if (m.getVarIndex () > varidx)
      return m;

    if (m.getVarIndex () < varidx)
      {
	MDDNode[] kids = new MDDNode [plicity];
	for (int i = 0; i < plicity; i++)
	  kids[i] = Restrict (m.getChild (i), varidx, value);

	return makeUnique (m.getVarIndex(), kids);		 
      }
    else // varidx == m.getVarIdx()
      //return Restrict (m.getChild (value), varidx, value);
      return m.getChild (value);
    // XXX ARIE: should not that be return m.getChild (value) ????
    // The two are, of course, equivalent... cdtw
	       
  }


  /***
   *** Performs existential quantification of 'node'
   *** 'cube' is a positive cube where we will only look at
   *** the 'top' branch of every node
   ***/
  public MDDNode existAbstract (MDDNode node, MDDNode cube)
  {

    // -- quick bail out
    if (cube.isConstant () || node.isConstant ()) return node;
    
    int top = s.idtimesIdx ();

    int largestIndex = cube.getLargestIndex (top);
    if (largestIndex < node.getVarIndex ())      
      return node;

    

    
    // -- top most (the lowest index) variable of the node
    int nodeIndex = node.getVarIndex ();

    // -- top most (the lowest index) variable of the cube
    int cubeIndex = cube.getVarIndex ();


    /** if the lowest variable in 'node' is higher than the one of the cube
     ** then the 'node' does not depend on this variable and quantification
     ** does not change anything, just skip an element of the cube and go on
     **/
    if (cubeIndex < nodeIndex)
      return existAbstract (node, cube.getChild (top));    
    

    // -- we only count operations that actually take time
    countOp (3);

    /*** we will have to do something from here so check the cache first */
    MDDNode result = null;
    if (caching)
      result = cache.find (node, cube, EXIST_ABSTRACT);

    if (result != null) 
      {
	//System.out.println ("existAbstract cache hit!");
	countCacheHit (3);
	return result;
      }
    result = getZero ();

    if (cubeIndex > nodeIndex)
      {
	// -- create a new node with 'nodeIndex' where each child is 
	// -- abstracted
	MDDNode[] children = new MDDNode [node.howManyChildren ()];
	for (int i = 0; i < children.length; i++)
	  children [i] = existAbstract (node.getChild (i), cube);
	result = makeUnique (nodeIndex, children);
      }
    else // -- cubeIndex == nodeIndex
      {
	int len = node.howManyChildren ();
	MDDNode nextCube = cube.getChild (top);
	// -- just compute all cofactors and produce a JOIN of them
	for (int i = 0; i < len; i++)
	  {
	    MDDNode child;
	    if (inIgnoreMap (i))
	      continue;
	    else
	      child = existAbstract (node.getChild (i), nextCube);	    
	    result = Apply (result, child, FiniteLattice.JOIN);
	    // -- we can bail out as soon as result is top
	    if (result == getOne ()) break;
	  }
	
      }

    // -- cache the result
    cache.insert (node, cube, EXIST_ABSTRACT, result);
    return result;
  }


  /*** Universal quantification */
  public MDDNode forallAbstract (MDDNode node, MDDNode cube)
  {

    // -- quick bail out
    if (cube.isConstant () || node.isConstant ()) return node;
    
    int top = s.idtimesIdx ();

    //      int largestIndex = cube.getLargestIndex (top);
    //      if (largestIndex < node.getVarIndex ())      
    //        return node;

    

    
    // -- top most (the lowest index) variable of the node
    int nodeIndex = node.getVarIndex ();

    // -- top most (the lowest index) variable of the cube
    int cubeIndex = cube.getVarIndex ();


    /** if the lowest variable in 'node' is higher than the one of the cube
     ** then the 'node' does not depend on this variable and quantification
     ** does not change anything, just skip an element of the cube and go on
     **/
    if (cubeIndex < nodeIndex)
      return forallAbstract (node, cube.getChild (top));
    
    // -- count operations that take time for stats
    countOp (4);

    /*** we will have to do something from here so check the cache first */
    MDDNode result = null;
    if (caching)
      result = cache.find (node, cube, FORALL_ABSTRACT);

    if (result != null) 
      {
	countCacheHit (4);
	return result;
      }

    if (cubeIndex > nodeIndex)
      {
	// -- create a new node with 'nodeIndex' where each child is 
	// -- abstracted
	MDDNode[] children = new MDDNode [node.howManyChildren ()];
	for (int i = 0; i < children.length; i++)
	  children [i] = forallAbstract (node.getChild (i), cube);
	result = makeUnique (nodeIndex, children);
      }
    else // -- cubeIndex == nodeIndex
      {
	result = getOne ();
	int len = node.howManyChildren ();
	MDDNode nextCube = cube.getChild (top);
	// -- just compute all cofactors and produce a MEET of them
	for (int i = 0; i < len; i++)
	  {
	    MDDNode child;
	    if (inIgnoreMap (i)) 
	      continue;
	    else
	      child = forallAbstract (node.getChild (i), nextCube);
	    result = Apply (result, child, FiniteLattice.MEET);
	    // -- we can bail out as soon as result is bot
	    if (result == getZero ()) break;
	  }
	
      }

    // -- cache the result
    cache.insert (node, cube, FORALL_ABSTRACT, result);
    
    return result;
  }

  boolean inIgnoreMap (int val)
  {
    if (ignoreMap == null) return false;
    for (int i = 0; i < ignoreMap.length; i++)
      if (ignoreMap [i] == val) return true;
    return false;
  }
  
  

  public MvState newState ()
  {
    return new MvState (plicity, s);
  }

  /** Iterate <i>oper</i> over the entire domain of a variable */
  public MDDNode BigOp (MDDNode m1, int varidx, int oper) {
    // XXX ARIE: This depends on the way Algebra represents logic values 
    // XXX ARIE: and thus should be re-done!
    MDDNode temp = Restrict(m1, varidx, 1);

    for (int i = 2; i < plicity; i++)
      temp = Apply (temp, Restrict (m1, varidx, i), oper);
    return temp;
  } // end MDDNode BigOp()

 

  // end of public methods
  // start of private methods
    
  /** Utility function. Mark all entries unvisited. */
  void markAllUnvisited () 
  {
    for (Iterator it = uniqueTable.values ().iterator (); it.hasNext ();)
      ((MDDNode)it.next ()).setVisited (false);
  } 

  public MDDNode renameVars (MDDNode mdd, int[] newVars)
  {
    // XXX This is only this simple because of the variable
    // XXX ordering we currently have. This is far from working 
    // XXX in general

    // -- we do not prime constant nodes
    if (mdd.isConstant ()) return mdd;

    MDDNode answer = null;
    answer = cache.find (mdd, RENAME_VARS, (Object)newVars);
    if (answer != null) return answer;

    // -- primes all of the variables of an MDDNode
    MDDNode[] renamedChildren = new MDDNode [mdd.howManyChildren ()];
    
    for (int i = 0; i < renamedChildren.length; i++)
      renamedChildren [i] = renameVars (mdd.getChild (i), newVars);

    int varIndex = mdd.getVarIndex ();    
    
    answer = makeUnique (newVars [varIndex], renamedChildren);
    cache.insert (mdd, RENAME_VARS, newVars, answer);
    return answer;
  }
  
  public int dagSize (MDDNode mddNode)
  {
    Set seen = new HashSet ();
    return dagSizeRecur (mddNode, seen);
  }
  
  public int sharedSize (MDDNode[] mddNodes)
  {
    Set seen = new HashSet ();
    int result = 0;
    for (int i = 0; i < mddNodes.length; i++)
      result += dagSizeRecur (mddNodes [i], seen);
    return result;
  }
  
  
  int dagSizeRecur (MDDNode node, Set seen)
  {
    // -- don't count the ones we've seen already
    if (seen.contains (node)) return 0;

    // -- we've seen this node so don't need to count it twice
    seen.add (node);

    // -- if node is constant don't have to go any further
    if (node.isConstant ()) return 1;

    int result = 1;
    // -- node is not constant so count the children and add one
    for (int i = 0; i < node.howManyChildren (); i++)
      result += dagSizeRecur (node.getChild (i), seen);

    return result;
  }


  /*** 
   *** Convert a DD to a new MDDManager 
   ***/
  public MDDNode transfer (MDDManager dstManager, MDDNode srcNode, 
			   int[] logicMap)
  {
    return transfer (dstManager, srcNode, logicMap, 1);
  }

  public MDDNode transfer (MDDManager dstManager, MDDNode srcNode, 
			   int[] logicMap, int factor)
  {
    Map transferCache = new HashMap ();
    return transfer (dstManager, srcNode, logicMap, factor, transferCache);
  }
  
  public MDDNode transfer (MDDManager dstManager, MDDNode srcNode, 
			   int[] logicMap, int factor, Map transferCache)
  {
    if (srcNode.isConstant ())
      return dstManager.getConstant (logicMap [srcNode.getValue ()]);
    
    MDDNode result = (MDDNode)transferCache.get (srcNode);
    if (result != null) return result;

    
    MDDNode[] children = new MDDNode [dstManager.getAlgebra ().size ()];
    
    Arrays.fill (children, dstManager.getZero ());
    
    for (int i = 0; i < srcNode.howManyChildren (); i++)
      {
	// -- skip if the map is not defined, this is useful to map
	// -- n-valued diagaram into m-valued diagram when m < n
	if (logicMap [i] == Algebra.NO_VALUE) continue;
	
	children [logicMap [i]] = transfer (dstManager, srcNode.getChild (i),
					    logicMap, factor, transferCache);
      }
    
    
    result = dstManager.makeUnique(srcNode.getVarIndex () * factor, 
				   children);
    transferCache.put (srcNode, result);
    return result;
  }
  

  public MvSet transfer (MvSetFactory dstFactory, MDDNode srcNode, 
			 int[] logicMap)
  {
    return transfer (dstFactory, srcNode, logicMap, 1);
  }

  public MvSet transfer (MvSetFactory dstFactory, MDDNode srcNode, 
			 int[] logicMap, int factor)
  {
    Map transferCache = new HashMap ();
    return transfer (dstFactory, srcNode, logicMap, factor, transferCache);
  }


  public MvSet transfer (MvSetFactory dstFactory, MDDNode srcNode, 
			 int[] logicMap, int factor, Map transferCache)
  {
    if (srcNode.isConstant ())
      return dstFactory.createConstant (logicMap [srcNode.getValue ()]);
    
    MvSet result = (MvSet)transferCache.get (srcNode);
    if (result != null) return result;

    
    MvSet[] children = new MvSet [dstFactory.getLattice ().size ()];
    
    Arrays.fill (children, 
		 dstFactory.createConstant 
		 (dstFactory.getLattice ().idplusIdx ()));
    
    for (int i = 0; i < srcNode.howManyChildren (); i++)
      {
	// -- skip if the map is not defined, this is useful to map
	// -- n-valued diagaram into m-valued diagram when m < n
	if (logicMap [i] == Algebra.NO_VALUE) continue;
	
	children [logicMap [i]] = transfer (dstFactory, srcNode.getChild (i),
					    logicMap, factor, transferCache);
      }
    
    
    result = 
      dstFactory.createCase (srcNode.getVarIndex () * factor, children);
    
    transferCache.put (srcNode, result);
    return result;
  }



  /***
   *** Reduces the DD by treating x = src as x = dst, for all x
   ***/
  public MDDNode reduce (MDDNode node, int src, int dst)
  {
    if (node.isConstant ())
      // -- nothing to do
      return node;
    

    // XXX Potentially we can do the merge and reduce at the same time
    // XXX but this should be good enough for now

    // -- array to hold children
    MDDNode[] children = new MDDNode [plicity];

    // -- recursively reduce the children
    for (int i = 0; i < children.length; i++)
      children [i] = reduce (node.getChild (i), src, dst);
    
    // -- merge children of src with dst
    children [dst] = 
      Apply (children [dst], children [src], FiniteLattice.JOIN);
    // -- remove src
    children [src] = getZero ();
    
    // -- done
    return makeUnique (node.getVarIndex (), children);
  }

  public MDDNode killEdge (MDDNode node, boolean[] map, int val)
  {
    if (node.isConstant ()) return node;
    
    MDDNode[] children = new MDDNode [plicity];
    
    for (int i = 0; i < children.length; i++)
      children [i] = killEdge (node.getChild (i), map, val);
    
    // -- for all boolean arguments, clean up 
    if (!map [node.getVarIndex ()])
      {
	// -- check if the node is constant except for val edge
	boolean constant = true;
	for (int i = 1; i < children.length; i++)
	  if (i != val && children [0] != children [i])
	    {
	      constant = false;
	      break;
	    }
	
	
	if (constant)
	  {
	    return children [0];
	  }
	
	else
	  children [val] = getZero ();
      }
    
    
    return makeUnique (node.getVarIndex (), children);
    
  }

  public MDDPoint getRandomPoint (MDDNode node, ChaosIterator varDomain)
  {
    // -- traverse the tree rooted at node, until a terminal node
    MDDPoint mddPoint = new MDDPoint (getOne (), getOne ().getValue ());
    
    return getRandomPointRecur (node, varDomain, mddPoint);
  }
  
  
  MDDPoint getRandomPointRecur (MDDNode node, ChaosIterator varDomain, 
				MDDPoint point)
  {
    if (node.isConstant ())
      {
	point.setValue (node.getValue ());
	return point;
      }
    

    // -- pick a child
    int childIdx = varDomain.nextInt ();

    // -- traverse the diagram first (otherwise we get an exp. algorithm)
    point = getRandomPointRecur (node.getChild (childIdx), 
				 varDomain, point);
    
    // -- now add this node to the top of the point generated so far
    MDDNode var = getVarEquals (node.getVarIndex (), childIdx);
    point.setNode (Apply (getVarEquals (node.getVarIndex (), childIdx),
			  point.getNode (),
			  FiniteLattice.MEET));
    return point;
    
  }
  

  public static interface ApplyFunction
  {
    // -- id to be used for caching
    int getId ();
    boolean isSymetric ();
    boolean isCacheable ();
    // -- apply the function to two arguments
    int binApply (int a, int b);
  }
  
  
}




