package edu.toronto.cs.mvset;

import edu.toronto.cs.mdd.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.algebra.*;
import java.util.Arrays;

/***
 *** Implementation of an MvSet based on the MDD library
 *** and join irreducible decomposition
 ***/
public class JBDDMvSet extends AbstractMvSet
{

  static final BooleanLattice boolLattice = new BooleanLattice ();

  JBDDMvSetFactory factory;
  
  // -- manager for the underlying MDDs
  MDDManager mddManager;

  // -- Finite lattice corresponding to the manager
  FiniteLattice lattice;
  
  // -- the actuall MDDs representing this function
  MDDNode[] mdd;
  // -- join irreducibles of the lattice such that
  // -- mdd [i] represents an l-cut w.r.t. joinIrr [i]
  int[] joinIrr;
  int[] invJoinIrr;
  
  // -- negation table
  int[] neg;
   
  public JBDDMvSet (JBDDMvSetFactory _factory, MDDNode[] _mdd)
  {
    mdd = _mdd;
    factory = _factory;
    mddManager = factory.getMddManager ();
    joinIrr = factory.getJoinIrr ();
    invJoinIrr = factory.getInvJoinIrr ();
    neg = factory.getNegTable ();
    lattice = factory.getLattice ();
  }
  




  /*** 
   *** Creates a pointwise composition
   *** op is an operator: L x L -> L
   *** f is the current mv-set
   *** result h (x) = f (x) op g (x)
   ***/
  public MvSet ptwiseCompose (int op, MvSet g)
  {
    MDDNode[] result = new MDDNode [joinIrr.length];

    MDDNode[] gSets = ((JBDDMvSet)g).getMddNodes ();
    
    for (int i = 0; i < result.length; i++)
      result [i] = mddManager.Apply (mdd [i], gSets [i], op);
    
    return createMvSet (result);
  }

  public MvSet ptwiseCompare (int op, MvSet g)
  {
    if (op != FiniteLattice.ABOVE &&
	op != FiniteLattice.BELOW &&
	op != FiniteLattice.EQUAL)
      Assert.assert ("Cannot comapre using: " + op);

    MDDNode node = mddManager.getOne ();
    
    MDDNode[] gSets = ((JBDDMvSet)g).getMddNodes ();

    for (int i = 0; i < gSets.length; i++)
      node = mddManager.Apply (node, 
			       mddManager.Apply (mdd [i], 
						 gSets [i], 
						 op), FiniteLattice.MEET);
    
    MDDNode[] result = new MDDNode [joinIrr.length];
    Arrays.fill (result, node);
    return createMvSet (result);
  }  

  public int size ()
  {
    return mddManager.sharedSize (mdd);
  }
  

  void dumpMvSet (String prefix)
  {
    for (int i = 0; i < mdd.length; i++)
      MDDNodeToDaVinci.mddToDaVinci (mdd [i], 
				     "/tmp/" + prefix + "_" + i + ".daVinci");
  }
  
  void dumpJoinIrrTable ()
  {
    for (int i = 0; i < joinIrr.length; i++)
      System.out.println ("#" + i + " = " + joinIrr [i]);
  }
  
 
  /***
   *** Creates a ptwise negation
   *** h (x) = \neg f (x)
   ***/
  public MvSet ptwiseNeg ()
  {
    MDDNode[] mddNodes = new MDDNode [joinIrr.length];
    
    // -- mddNodes [i] is the result of applying a lattice NEG
    // -- to the mddNodes [j] where j corresponds to the neg table value
    // -- of join irreducible corresponding to mdd [i]
    for (int i = 0; i < joinIrr.length; i++)
      mddNodes [i] = mddManager.Apply (mdd [invJoinIrr [neg [joinIrr [i]]]],
					 BooleanLattice.NEG);
    return createMvSet (mddNodes);
  }
  


  /***
   *** Restricts an argument
   *** h (x_0, x_1, x_2, ...) = f (x_0, ..., value, ...)
   *** where value is substituted at argIdx
   ***/
  public MvSet restrict (int argIdx, int value)
  {
    MDDNode[] result = new MDDNode [joinIrr.length];
    
    System.arraycopy (mdd, 0, result, 0, mdd.length);
    
    for (int i = 0; i < result.length; i++)
      {
	// -- restrict each argIdx_k variable (k ranges over join irr)
	int startIdx = argIdx * joinIrr.length;
	for (int k = 0; k < joinIrr.length; k++)
	  {
	    int bddValue = lattice.leq (joinIrr [k], value) ?
	      boolLattice.idtimesIdx () : boolLattice.idplusIdx ();
	    result [i] = 
	      mddManager.Restrict (result [i], startIdx + k, bddValue);
	  }	
      }
    
    return createMvSet (result);
  }
  
  public MvSet restrict (MvState state)
  {
    MDDNode[] result = new MDDNode [joinIrr.length];

    System.arraycopy (mdd, 0, result, 0, mdd.length);
    
    // -- for each bdd 
    // -- for each variable
    // -- for each variable_k corresponding to a join irr k
    // -- restrict
    for (int i = 0; i < result.length; i++)
      for (int j = state.getFirstVar (); j <= state.getLastVar (); j++)
	{
	  if (!state.isSet (j))
	      continue;


	  int startIdx = j * joinIrr.length;
	  
	  for (int k = 0; k < joinIrr.length; k++)
	    {
	      int bddValue =
		lattice.leq (joinIrr [k], state.getValue (j)) ?
		boolLattice.idtimesIdx () : boolLattice.idplusIdx ();


	      result [i] = 
		mddManager.Restrict (result [i], startIdx + k, bddValue);
	    }
	}
    
    return createMvSet (result);
  }
  

  /***
   *** Renames the arguments. newArgs is a map from old args to new so that
   *** h (x) = f (newArgs [0], newArgs [1], ...)
   ***/
  public MvSet renameArgs (int[] _newArgs)
  {

    int[] newArgs = new int [_newArgs.length * joinIrr.length];

    for (int i = 0; i < _newArgs.length; i++)
      {
	int oldVar = i * joinIrr.length;
	int newVar = _newArgs [i] * joinIrr.length;
	
	for (int k = 0; k < joinIrr.length; k++)
	  newArgs [oldVar + k] = newVar + k;	
      }

    MDDNode[] result = new MDDNode [joinIrr.length];
    for (int i = 0; i < result.length; i++)
      result [i] = mddManager.renameVars (mdd [i], newArgs);

    return createMvSet (result);
  }
  
  public MvSet existAbstract (MvSet cube)
  {
    return createMvSet (existAbstract (((JBDDMvSet)cube).getMddNodes ()[0]));
  }
  
  public MvSet existAbstract (int[] varIndex)
  {
    int top = lattice.idtimesIdx ();
    MDDNode cube = mddManager.getOne ();
    for (int i = 0; i < varIndex.length; i++)
      cube = mddManager.Apply (cube, 
			       factory.createBDDForVar (varIndex [i], top),
			       BooleanLattice.MEET);
    return createMvSet (existAbstract (cube));
  }
  
  MDDNode[] existAbstract (MDDNode cube)
  {
    MDDNode[] bdds = new MDDNode [joinIrr.length];
    
    for (int i = 0; i < bdds.length; i++)
      bdds [i] = mddManager.existAbstract (mdd [i], cube);
    return bdds;
  }
  

  public MvSet forallAbstract (MvSet cube)
  {
    return createMvSet (forallAbstract (((JBDDMvSet)cube).getMddNodes ()[0]));
  }
  
  public MvSet forallAbstract (int[] varIndex)
  {
    int top = lattice.idtimesIdx ();
    MDDNode cube = mddManager.getOne ();
    for (int i = 0; i < varIndex.length; i++)
      cube = mddManager.Apply (cube, 
			       factory.createBDDForVar (varIndex [i], top),
			       BooleanLattice.MEET);
    return createMvSet (forallAbstract (cube));
  }
  
  MDDNode[] forallAbstract (MDDNode cube)
  {
    MDDNode[] bdds = new MDDNode [joinIrr.length];
    
    for (int i = 0; i < bdds.length; i++)
      bdds [i] = mddManager.forallAbstract (mdd [i], cube);
    return bdds;
  }

  
  /***
   *** Evaluates this function on an input
   *** result = f (values [0], values [1], ...)
   ***/
  public int evaluate (int[] values)
  {
    return evaluate (new MvState (values, lattice));
  }

  public int evaluate (MvState state)
  {
    MvSet restricted = this.restrict (state);

    MDDNode[] mddNodes = ((JBDDMvSet)restricted).getMddNodes ();
    
    int value = lattice.idplusIdx ();
    
    for (int i = 0; i < mddNodes.length; i++)
      if (mddNodes [i] == mddManager.getOne ())
	value = lattice.plus (value, joinIrr [i]);
    return value;
  }
  


  /*** Arrity of this function */
  public int getArrity ()
  {
    return factory.getArrity ();
  }

  /*** Gets the lattice */
  public FiniteLattice getLattice ()
  {
    return lattice;
  }

  public MDDNode[] getMddNodes ()
  {
    return mdd;
  }

  protected MvSetFactory getMvSetFactory ()
  {
    return factory;
  }
  
  private MvSet createMvSet (MDDNode[] mddNode)
  {
    return factory.createMvSet (mddNode);
  }
  

  public boolean equals (Object o)
  {
    if (! (o instanceof JBDDMvSet)) return false;
    return equals ((JBDDMvSet)o);
  }
  
  public boolean equals (JBDDMvSet v)
  {
    // -- two mv-sets can only be equal if they come from the same
    // -- factory
    if (factory != v.getMvSetFactory ())
      return false;

    // -- two mv-sets are equal if their mdd representations are equal
    MDDNode[] gSets = v.getMddNodes ();
    for (int i = 0; i < mdd.length; i++)
      if (! (mdd [i].equals (gSets [i]))) return false;
    return true;    
  }
  

  public static MvSetFactory newMvSetFactory (FiniteLattice lattice, 
					      int arrity)
  {
    return new JBDDMvSetFactory (lattice, arrity);
  }

  public MvSetFactory getFactory ()
  {
    return factory;
  }  

  public static class JBDDMvSetFactory extends AbstractMvSetFactory
  {
    int arrity;
    MDDManager mddManager;
    FiniteLattice lattice;

    int[] joinIrr;
    // -- this is the inverse of joinIrr
    // -- let j be a join irreducible 
    // -- invJoinIrr [j] == i -> joinIrr [i] == j
    // -- it's size is the size of the lattice, 
    // -- and it is not defined for non-join irreducible elements
    int[] invJoinIrr;
    int[] neg;
    
    public JBDDMvSetFactory (FiniteLattice _lattice,
			      int _arrity)
    {
      arrity = _arrity;
      // XXX BIG HACK actually better to create boolean lattice and 
      // XXX boolean mdd here
      arrity = 2;
      lattice = _lattice;
      mddManager = new MDDMgr (arrity, JBDDMvSet.boolLattice);
      joinIrr = lattice.getJoinIrreducibles ();

      invJoinIrr = computeInvJoinIrr (lattice.size (), joinIrr);
      
      // -- create the negation table
      neg = new int [lattice.size ()];
      computeNegTable ();
      
    }
    
    private int[] computeInvJoinIrr (int size, int[] joinIrr)
    {
      int[] invJoinIrr = new int [size];
      for (int i = 0; i < joinIrr.length; i++)
	invJoinIrr [joinIrr [i]] = i;
      return invJoinIrr;
    }
    

    /***
     *** Given a value in L creates a function 
     *** f(x_0, x_1, ...) = value
     ***/
    public MvSet createConstant (int value)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];

      Arrays.fill (mddNodes, mddManager.getZero ());
      
      int[] decomp = lattice.joinDecomposition (value);
      for (int i = 0; i < decomp.length; i++)
	mddNodes [invJoinIrr [decomp [i]]] = mddManager.getOne ();

      return createMvSet (mddNodes);
    }    
    
    /***
     *** Given an argument index (0 <= argIdx < n)
     *** constructs a projection function
     *** f(x_0, x_1, ...) = x_argIdx
     ***/
    public MvSet createProjection (int argIdx)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];

      // -- each variable x is now encoded as 
      // -- x_0, x_1, x_2, ... x_n where n is number of join irreducibles
      // -- where x_i is true iff join irreducible i is bellow the value 
      // -- of x
      int startIndex = argIdx * joinIrr.length;
      
      for (int i = 0; i < joinIrr.length; i++)
	mddNodes [i] = 
	  mddManager.getVarEquals (startIndex + i, 
				   JBDDMvSet.boolLattice.idtimesIdx ());
      
      
      return createMvSet (mddNodes);
    }    
    
    
    /***
     *** Creates a point function 
     *** f (args) = value
     ***          = 0 otherwise
     ***/
    public MvSet createPoint (int[] args, int value)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];

      MDDNode mddNode = mddManager.getOne ();
      for (int i = 0; i < args.length; i++)
	{
	  if (args [i] == Algebra.NO_VALUE) continue;
	  mddNode = mddManager.Apply (mddNode, 
				      createBDDForVar (i, args [i]), 
				      BooleanLattice.MEET);
	}
      

      for (int i = 0; i < mddNodes.length; i++)
	{
	  if (!lattice.leq (joinIrr [i], value))
	    mddNodes [i] = mddManager.getZero ();
	  else
	    mddNodes [i] = mddNode;
	}

     return createMvSet (mddNodes);
    }

    
    // -- create a BDD representation for a variable 
    MDDNode createBDDForVar (int var, int val)
    {
      // -- decompose var into joinIrr components, 
      // -- a component is true if the join irreducible is bellow val

      int startIdx = var * joinIrr.length;
      
      MDDNode mddNode = mddManager.getOne ();
      
      for (int i = 0; i < joinIrr.length; i++)
	{
	  int value = lattice.leq (joinIrr [i], val) ? 
	    boolLattice.idtimesIdx () : 
	    boolLattice.idplusIdx ();

	  mddNode = 
	    mddManager.Apply (mddNode, 
			      mddManager.getVarEquals (startIdx + i, value), 
			      BooleanLattice.MEET);
	}

      return mddNode;
    }
    

    public MvSet var (int argIdx, int argVal, int value)
    {
      MDDNode mddNode = createBDDForVar (argIdx, argVal);
      
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];
      
      for (int i = 0; i < mddNodes.length; i++)
	{
	  if (getLattice ().leq (joinIrr [i], value))
	    mddNodes [i] = mddNode;
	  else
	    mddNodes [i] = mddManager.getZero ();
	}
      return createMvSet (mddNodes);
    }
    
    public MvSet createPoint (MvState state, int value)
    {
      MDDNode[] mddNodes = new MDDNode [joinIrr.length];
      
      MDDNode mddNode = mddManager.getOne ();
      for (int i = state.getFirstVar (); i <= state.getLastVar (); i++)
	{
	  if (!state.isSet (i)) continue;
	  
	  mddNode = 
	    mddManager.Apply (mddNode, 
			      createBDDForVar (i, state.getValue (i)), 
			      BooleanLattice.MEET);
	}
      
      for (int i = 0; i < joinIrr.length; i++)
	{
	  if (!lattice.leq (joinIrr [i], value))
	    mddNodes [i] = mddManager.getZero ();
	  else
	    mddNodes [i] = mddNode;
	}
      
      return createMvSet (mddNodes);
    }


    public MvSet createMvSet (MDDNode[] mddNodes)
    {
      return new JBDDMvSet (this, mddNodes);
    }

    public FiniteLattice getLattice ()
    {
      return lattice;
    }
    
    public  MDDManager getMddManager ()
    {
      return mddManager;
    }
    
    protected int getArrity ()
    {
      return arrity;
    }
    
    protected int[] getJoinIrr ()
    {
      return joinIrr;
    }
    protected int[] getInvJoinIrr ()
    {
      return invJoinIrr;
    }
    
    protected int[] getNegTable ()
    {
      return neg;
    }

    private void computeNegTable ()
    {
      int[] f = computeF ();
      
      for (int i = 0; i < joinIrr.length; i++)
	{
	  int j = joinIrr [i];
	  neg [lattice.unOp (f [j], FiniteLattice.NEG)] = j;
	}      
    }
    private int[] computeF ()
    {
      int[] f = new int [lattice.size ()];
      
      for (int i = 0; i < joinIrr.length; i++)
	f [joinIrr [i]] = computeF (joinIrr [i]);
      return f;
    }
    private int computeF (int j)
    {
      // -- return max (L - upSet (j))
      int[] upSet = lattice.getUpSet (j);
      
      int result = lattice.idplusIdx ();
      
      for (IntIterator it = lattice.carrierSet (); it.hasNext ();)
	{
	  int x = it.nextInt ();
	  if (lattice.eq (x, j)) continue;
	  // x < j ==> x \in (L - up (j))
	  else if (lattice.leq (x, j)) 
	    result = lattice.plus (result, x);
	  // x incomparable to j ==> x \in (L - up (j))
	  else if (!lattice.geq (x, j)) 
	    result = lattice.plus (result, x);
	}
      
      return result;
    }
    
    
    
  }
  

}





  
  
