package edu.toronto.cs.smv;

import java.util.*;
import edu.toronto.cs.modelchecker.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.ctl.*;

/*** Class representing an SMV module */
public abstract class SMVModule
{
  // -- variable separator
  public static final String VAR_SEP = "_";
  public static final String NEXT_SUFFIX = "\'";
  // -- the name of the module
  String name;
  String modelName = null;
  
  // -- all declared variables
  protected String[] vars;

  // -- precomputed named variables
  protected String[] varsNamed = null;

  // -- map between variable names and their index
  Map varIndex;

  // -- factory to create MvSets
  protected MvSetFactory ddManager = null;
  // -- lattice for MvSets
  protected FiniteLattice lattice = null;
  protected int bot;
  protected int top;
  protected PropOrderIf ordering = null;

  protected MvSet mvBot;
  protected MvSet mvTop;
  
  public SMVModule (String _name)
  {
    name = _name;
    vars = new String[0];
  }


  // -- enumerated types that we know about
  public EnumeratedTypes getEnumTypes ()
  {
    // -- we have none here
    return null;
  }
  
  
  public String getName ()
  {
    return name;
  }
  
  public String getModelName ()
  {
    return modelName;
  }
  protected void setModelName (String _modelName)
  {
    modelName = _modelName;
  }
  
  public void setMvSetFactory (MvSetFactory _ddManager)
  {
    ddManager = _ddManager;
    mvTop = ddManager.createConstant (top);
    mvBot = ddManager.createConstant (bot);
  }

  public void setLattice (FiniteLattice _lattice)
  {
    lattice = _lattice;
    top = lattice.idtimesIdx ();
    bot = lattice.idplusIdx ();
  }
  

  public FiniteLattice getLattice ()
  {
    return lattice;
  }
  
    
  public String[] getVars ()
  {
    if (varsNamed == null)
      {
	varsNamed = new String [vars.length];
	for (int i = 0; i < vars.length; i++)
	  varsNamed [i] = makeGlobalVarName (vars [i]);
	
      }
    return varsNamed;
  }

  String makeGlobalVarName (String varName)
  {
    return getName () == null || getName ().length () == 0 ?
      varName : getName () + VAR_SEP + varName;
  }
  


  public int getVarIndex (String varName)
  {
    return ((Integer)varIndex.get (makeGlobalVarName (varName))).intValue ();
  }
  

  public List getSpecs ()
  {
    return Collections.EMPTY_LIST;
  }
  

  // -- set variable ordering
  public void setVarOrdering (PropOrderIf _ordering)
  {
    System.out.println("SMVM: setVarOrdering called with vars.length ="+
		       (varsNamed.length));
    
    
    ordering = _ordering;
    varIndex = new HashMap ();
    
    for (int i = 0; i < vars.length; i++)
      {
	System.out.println("Adding variable "+varsNamed[i]);
	
	varIndex.put (varsNamed [i], 
		      new Integer (ordering.getPropOrder (varsNamed [i])));
	varIndex.put 
	  (varsNamed [i] + NEXT_SUFFIX,
	   new Integer (ordering.getNextPropOrder (varsNamed [i])));	
      }
  }
  
  // -- given an index of the variable in the 'vars' array, return its
  // -- order in the Propositional Ordering given above
  public int getVarOrder (int idx)
  {
    return ((Integer)varIndex.get (varsNamed [idx])).intValue ();
  }
  
  public int getNextVarOrder (int idx)
  {
    return
      ((Integer)varIndex.get (varsNamed [idx] + NEXT_SUFFIX)).intValue ();

  }


  
  // -- computes an MvSet representation of the transition relation
  public abstract MvSet computeTrans ();
  public abstract MvState initializeInitState (MvState initState);

  public MvSet cur (int idx)
  {
    return cur (idx, lattice.idtimesIdx (), lattice.idtimesIdx ());
  }

  public MvSet cur (int idx, int val)
  {
    return cur (idx, val, lattice.idtimesIdx ());
  }

  public MvSet cur (int varIdx, int varVal, int val)
  {
    return ddManager.var (getVarOrder (varIdx), varVal, val);
  }
  
  
  public MvSet next (int idx)
  {
    return next (idx, lattice.idtimesIdx (), lattice.idtimesIdx ());
  }

  public MvSet next (int idx, int val)
  {
    return next (idx, val, lattice.idtimesIdx ());
  }
  
  public MvSet next (int varIdx, int varVal, int val)
  {
    return ddManager.var (getNextVarOrder (varIdx), varVal, val);
  }

  // -- computes a decision diagram for a variable 'x' of type 'nat'
  // -- corresponding to 'x = val'
  // -- baseIdx -- the base index of the boolean variables representing 'x'
  // -- width -- number of bits in 'x'
  // -- val -- is the value we are interested in
  public MvSet computeNatEquals (int baseIdx, int width, int val)
  {
    MvSet result = ddManager.createConstant (top);
    
    for (int i = 0; i < width; i++)
      result = result.and (cur (baseIdx + i, isBitSet (val, i) ? top : bot));
    return result;
  }
  public MvSet computeNextNatEquals (int baseIdx, int width, int val)
  {
  MvSet result = ddManager.createConstant (top);
    
    for (int i = 0; i < width; i++)
      result = result.and (next (baseIdx + i, isBitSet (val, i) ? top : bot));
    return result;
  }
  
  public MvSet computeNatLess (int baseIdx, int width, int val)
  {
    MvSet result = mvBot;
    
    for (int i = 0; i < val; i++)
      result = result.or (computeNatEquals (baseIdx, width, i));
    return result;
  }
  public MvSet computeNatGreater (int baseIdx, int width, int val, int topVal)
  {
    MvSet result = mvBot;
    
    for (int i = topVal; i > val; i--)
      result = result.or (computeNatEquals (baseIdx, width, i));
    return result;
  }
  
  
  
  // -- returns true if bit 'bit' of 'val' is set, false otherwise
  boolean isBitSet (int val, int bit)
  {
    return ((val >> bit) & 1) > 0;
  }
  

  public static class CaseStatement
  {

    // -- final result
    MvSet result = null;
    // -- conjunction of negations of prev. conditions
    // -- i.e. !cond1 & !cond2 & !cond3 ... etc
    MvSet condNeg = null;


    public CaseStatement ()
    {
    }

    // -- add one case of the form
    //  cond: stmt
    public CaseStatement addCase (MvSet cond, MvSet stmt)
    {
      if (result == null)
	{
	  result = cond.and (stmt);
	  condNeg = cond.not ();
	}
      else
	{
	  result = result.or (condNeg.and (cond.and (stmt)));
	  condNeg = condNeg.and (cond.not ());
	}
      return this;
    }

    // -- add the default case
    // 1 : stmt
    public CaseStatement addDefault (MvSet stmt)
    {
      result = result.or (condNeg.and (stmt));
      condNeg = null;
      return this;
    }
    
    public MvSet compute ()
    {
      return result;
    }    
    
  }

  public static class BoolCaseStatement
  {

    // -- final result
    MvSet result = null;
    // -- conjunction of negations of prev. conditions
    // -- i.e. !cond1 & !cond2 & !cond3 ... etc
    MvSet condNeg = null;
    MvSet mvBot;

    public BoolCaseStatement (MvSet _mvBot)
    {
      mvBot = _mvBot;
    }

    // -- add one case of the form
    //  cond: stmt
    public BoolCaseStatement addCase (MvSet cond, MvSet stmt)
    {
      if (result == null)
	{
	  result = cond.and (stmt);
	  condNeg = cond.eq (mvBot);
	}
      else
	{
	  result = result.or (condNeg.and (cond.and (stmt)));
	  condNeg = condNeg.and (cond.eq (mvBot));
	}
      return this;
    }

    // -- add the default case
    // 1 : stmt
    public BoolCaseStatement addDefault (MvSet stmt)
    {
      result = result.or (condNeg.and (stmt));
      condNeg = null;
      return this;
    }
    
    public MvSet compute ()
    {
      return result;
    }    
    
  }

  // -- an SMV SPEC clause
  public static interface SMVSpec
  {
    // -- returns the string representation of the spec
    String toString ();
    // -- the CTL representation of the spec
    CTLNode getCTL ();
  }
  
}














