package edu.toronto.cs.xkripke;

import edu.toronto.cs.modelchecker.*;
import edu.toronto.cs.xkripke.XKripke.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.mvset.*;

import java.util.Iterator;

// -- given a XKripke model creates an MDD for it
public class XKripkeMvSetBuilder
{

  /**
   ** Given an array of xkripke structures as well as the MvSetFactory
   ** and the propositional ordering constructs the MvSet representing
   ** the transition relation of the product of all the xkripke
   ** structures.
   **
   ** WARNING:
   **          1) assumes that all the xkripke structures use the same
   **             lattice
   **          2) assumes that all the xkripke structures use the same
   **             propositional ordering
   **/
  public static MvSet buildMvSet (MvSetFactory mvSetFactory,
				  XKripke [] kripke, 
				  PropOrderIf propOrder)
  {
    // get the lattice
    FiniteLattice lattice = kripke [0].getLattice ();
    // the resulting MvSet
    MvSet finalresult;

    // initialise the finalresult to the identity on MEET
    finalresult = mvSetFactory.createConstant (lattice.idtimesIdx ());
    // take all the xkripke structures one at a time and find the
    // product of the MvSet for them and the finalresult
    for (int x = 0; x < kripke.length; x++)
      {
	MvSet newset = buildMvSet (mvSetFactory, kripke [x], propOrder);
	finalresult = finalresult.ptwiseCompose (FiniteLattice.MEET, newset);
      }
    
    return finalresult;
  }
  /***
   *** mvSetFactory -- needed to create MvSets
   *** kripke -- the XKripke model to convert
   *** propOrder -- interface to order the atomic props
   ***/
  public static MvSet buildMvSet (MvSetFactory mvSetFactory, XKripke kripke, 
				  PropOrderIf propOrder)
  {
    // -- we build each transition individually and then 
    // -- join all of them together
    
    FiniteLattice lattice = kripke.getLattice ();
    // -- start with the identity on JOIN
    MvSet result = mvSetFactory.createConstant (lattice.idplusIdx ());

    for (Iterator it = kripke.getTransitions ().iterator (); it.hasNext ();)
      result = result.ptwiseCompose 
	(FiniteLattice.JOIN,
	 buildTransition (mvSetFactory, lattice, 
			  (XKripkeTransition)it.next (), propOrder));
    return result;
  }

  public static MvSet buildTransition (MvSetFactory mvSetFactory, 
				       FiniteLattice lattice,
				       XKripkeTransition trans,
				       PropOrderIf propOrder)
  {
    int idx = 0; // -- variable indexing
    
    MvState state = new MvState (propOrder.size (), lattice);

    for (Iterator it = propOrder.orderTransition (trans); it.hasNext ();)
      {
	XKripkeProp prop = (XKripkeProp)it.next ();
	// some propositions don't exist in this xkripke structure
	if (prop != null)
	  {
	    int latticeValue = lattice.getElementIdxByKey (prop.getValue ());
	    state.setValue (idx, latticeValue);
	  }
	idx++;
      }
    int transitionValue = lattice.getElementIdxByKey (trans.getValue ());
    MvSet r = mvSetFactory.createPoint (state, transitionValue);
    return r;
  }


  public static void main (String[] args) throws Exception 
  {
    XKripke kripke = XKripkeFactory.parse (args [0]);
    PropOrderIf propOrdering = 
      new AlternatePropOrder (kripke.getPropNames ());

    

    MvSetFactory mvSetFactory = 
      MDDMvSet.newMvSetFactory (kripke.getLattice (), propOrdering.size ());
    
    MvSet mvSet = buildMvSet (mvSetFactory, kripke, propOrdering);
    MDDNode mdd = ((MDDMvSet)mvSet).getMddNode ();
    System.out.println ("Is final mdd constant: " + mdd.isConstant ());
    System.out.println ("Writing mdd into systemMdd.daVinci");
    MDDNodeToDaVinci.mddToDaVinci (mdd, propOrdering, "systemMdd.daVinci");
    System.out.println ("Property ordering is as follows:");
    
    int size = propOrdering.size ();
    for (int i = 0; i < size; i++)
      {
	System.out.print ("#" + i + " = " + propOrdering.getPropName (i));
	if (propOrdering.isPostImage (i)) 
	  System.out.print ("'");
	System.out.println ();	
      }
    
    
  }
  
  
}



